# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================


import numpy as np

from magnopy._parameters._p22 import from_iso
from magnopy._spinham._convention import Convention
from magnopy._spinham._hamiltonian import SpinHamiltonian

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def cubic_ferro_nn(
    a: float = 1,
    J_iso: float = 1,
    J_21=(0, 0, 0),
    S: float = 0.5,
    dimensions: int = 3,
) -> SpinHamiltonian:
    r"""
    Prepares ferromagnetic Hamiltonian on the cubic lattice with one atom per unit cell.

    Only nearest-neighbor isotropic exchange interactions and on-site quadratic
    anisotropy are populated. The Hamiltonian has the form

    .. math::

        \mathcal{H}
        =
        \dfrac{1}{2}
        \sum_{\mu, \nu}
        J_{iso}
        \boldsymbol{S}_{\mu}
        \cdot
        \boldsymbol{S}_{\mu+\nu}
        +
        \sum_{\mu}
        \boldsymbol{S}_{\mu}
        \boldsymbol{J}_{21}
        \boldsymbol{S}_{\mu}

    where

    * :math:`\nu \in {(1,0,0), (-1,0,0)}` if ``dimensions == 1``.
    * :math:`\nu \in {(1,0,0), (-1,0,0), (0,1,0), (0,-1,0)}` if ``dimensions == 2``.
    * :math:`\nu \in {(1,0,0), (-1,0,0), (0,1,0), (0,-1,0), (0,0,1), (0,0,-1)}` if
      ``dimensions == 3``.

    Parameters
    ----------

    a : float, default 1.0
        Lattice parameter of the cubic lattice.
        The unit cell of the Hamiltonian is defined as

        .. code-block:: python

            [[a, 0, 0], [0, a, 0], [0, 0, a]]

    J_iso : float, default 1.0
        Isotropic exchange parameter for the pairs of the nearest-neighbor magnetic
        sites, given in energy units. Only magnitude is important, sign is ignored.

    J_21 : (3, 3) or (3, ) |array-like|_, default (0, 0, 0)
        On-site quadratic anisotropy.

    S : float, default 0.5
        Spin value of the magnetic site.

    dimensions : int, default 3
        Either 1, 2 or 3. Dimensionality of the spin Hamiltonian.

    Returns
    -------

    spinham : :py:class:`.SpinHamiltonian`
        Spin Hamiltonian.

    Examples
    --------

    To get an example with the default values use

    .. doctest::

        >>> import magnopy
        >>> spinham = magnopy.examples.cubic_ferro_nn()
        >>> spinham.cell
        array([[1., 0., 0.],
               [0., 1., 0.],
               [0., 0., 1.]])
        >>> spinham.atoms.names
        ['X']
        >>> spinham.atoms.spins
        [0.5]
        >>> spinham.atoms.positions
        [[0, 0, 0]]
        >>> for alpha, parameter in spinham.p21:
        ...     print(alpha, parameter, sep="\n")
        0
        [[0. 0. 0.]
         [0. 0. 0.]
         [0. 0. 0.]]
        >>> for alpha, beta, nu, parameter in spinham.p22:
        ...     print(alpha, beta, nu)
        ...     print(parameter)
        0 0 (0, 0, 1)
        [[-1. -0. -0.]
         [-0. -1. -0.]
         [-0. -0. -1.]]
        0 0 (0, 1, 0)
        [[-1. -0. -0.]
         [-0. -1. -0.]
         [-0. -0. -1.]]
        0 0 (1, 0, 0)
        [[-1. -0. -0.]
         [-0. -1. -0.]
         [-0. -0. -1.]]
        0 0 (0, 0, -1)
        [[-1. -0. -0.]
         [-0. -1. -0.]
         [-0. -0. -1.]]
        0 0 (0, -1, 0)
        [[-1. -0. -0.]
         [-0. -1. -0.]
         [-0. -0. -1.]]
        0 0 (-1, 0, 0)
        [[-1. -0. -0.]
         [-0. -1. -0.]
         [-0. -0. -1.]]

    With this function one can customize a few things of the Hamiltonian:

    *   Lattice parameter

        .. doctest::

            >>> import magnopy
            >>> spinham = magnopy.examples.cubic_ferro_nn(a=2)
            >>> spinham.cell
            array([[2., 0., 0.],
                   [0., 2., 0.],
                   [0., 0., 2.]])

    *   Spin values

        .. doctest::

            >>> import magnopy
            >>> spinham = magnopy.examples.cubic_ferro_nn(S=1.5)
            >>> spinham.atoms.spins
            [1.5]


    *   Value of the isotropic exchange

        .. doctest::

            >>> import magnopy
            >>> spinham = magnopy.examples.cubic_ferro_nn(J_iso=2)
            >>> for alpha, beta, nu, parameter in spinham.p22:
            ...     print(alpha, beta, nu)
            ...     print(parameter)
            0 0 (0, 0, 1)
            [[-2. -0. -0.]
             [-0. -2. -0.]
             [-0. -0. -2.]]
            0 0 (0, 1, 0)
            [[-2. -0. -0.]
             [-0. -2. -0.]
             [-0. -0. -2.]]
            0 0 (1, 0, 0)
            [[-2. -0. -0.]
             [-0. -2. -0.]
             [-0. -0. -2.]]
            0 0 (0, 0, -1)
            [[-2. -0. -0.]
             [-0. -2. -0.]
             [-0. -0. -2.]]
            0 0 (0, -1, 0)
            [[-2. -0. -0.]
             [-0. -2. -0.]
             [-0. -0. -2.]]
            0 0 (-1, 0, 0)
            [[-2. -0. -0.]
             [-0. -2. -0.]
             [-0. -0. -2.]]

    *   Diagonal of the on-site quadratic anisotropy

        .. doctest::

            >>> import magnopy
            >>> spinham = magnopy.examples.cubic_ferro_nn(J_21=(1, 2, -1))
            >>> for alpha, parameter in spinham.p21:
            ...     print(alpha, parameter, sep="\n")
            0
            [[ 1.  0.  0.]
             [ 0.  2.  0.]
             [ 0.  0. -1.]]

    *   Dimensionality of the nearest neighbors

        .. doctest::

            >>> import magnopy
            >>> spinham = magnopy.examples.cubic_ferro_nn(dimensions=1)
            >>> for alpha, beta, nu, parameter in spinham.p22:
            ...     print(alpha, beta, nu)
            ...     print(parameter)
            0 0 (1, 0, 0)
            [[-1. -0. -0.]
             [-0. -1. -0.]
             [-0. -0. -1.]]
            0 0 (-1, 0, 0)
            [[-1. -0. -0.]
             [-0. -1. -0.]
             [-0. -0. -1.]]
            >>> spinham = magnopy.examples.cubic_ferro_nn(dimensions=2)
            >>> for alpha, beta, nu, parameter in spinham.p22:
            ...     print(alpha, beta, nu)
            ...     print(parameter)
            0 0 (0, 1, 0)
            [[-1. -0. -0.]
             [-0. -1. -0.]
             [-0. -0. -1.]]
            0 0 (1, 0, 0)
            [[-1. -0. -0.]
             [-0. -1. -0.]
             [-0. -0. -1.]]
            0 0 (0, -1, 0)
            [[-1. -0. -0.]
             [-0. -1. -0.]
             [-0. -0. -1.]]
            0 0 (-1, 0, 0)
            [[-1. -0. -0.]
             [-0. -1. -0.]
             [-0. -0. -1.]]

    """

    cell = a * np.eye(3, dtype=float)
    atoms = dict(
        names=["X"],
        species=["X"],
        spins=[S],
        g_factors=[2],
        positions=[[0, 0, 0]],
    )
    convention = Convention(
        spin_normalized=False,
        multiple_counting=True,
        c21=1,
        c22=0.5,
    )

    spinham = SpinHamiltonian(cell=cell, atoms=atoms, convention=convention)

    J_21 = np.array(J_21, dtype=float)
    if J_21.shape == (3,):
        J_21 = np.diag(J_21)

    spinham.add_21(alpha=0, parameter=J_21)

    J_iso = -from_iso(iso=abs(J_iso))

    spinham.add_22(alpha=0, beta=0, nu=(1, 0, 0), parameter=J_iso)

    if dimensions >= 2:
        spinham.add_22(alpha=0, beta=0, nu=(0, 1, 0), parameter=J_iso)

    if dimensions >= 3:
        spinham.add_22(alpha=0, beta=0, nu=(0, 0, 1), parameter=J_iso)

    return spinham


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================
import random
import numpy as np
from magnopy._plotly_engine import PlotlyEngine
from wulfric.crystal import get_vector
from magnopy._parameters._p22 import to_dmi, to_iso
from magnopy._spinham._hamiltonian import SpinHamiltonian

try:
    import matplotlib.pyplot as plt

    MATPLOTLIB_AVAILABLE = True
    MATPLOTLIB_ERROR_MESSAGE = "If you see this message, please contact developers of the code (see magnopy.org)."
except ImportError:
    MATPLOTLIB_AVAILABLE = False
    MATPLOTLIB_ERROR_MESSAGE = "\n".join(
        [
            "Installation of matplotlib is not found, can not produce .png files",
            "Please install matplotlib with",
            "",
            "    pip install matplotlib",
            "",
        ]
    )

from magnopy._constants._units import _MAGNON_ENERGY_UNITS


def plot_spinham(
    spinham: SpinHamiltonian,
    distance_digits=5,
    plot_dmi=True,
    dmi_vectors_scale=1,
    _sphinx_gallery_fix=False,
):
    r"""
    Visualizes spin Hamiltonian.

    .. warning::

        Experimental feature. Only 1, 21, 22 parameters are implemented.

    .. versionadded:: 0.2.0

    Parameters
    ----------
    spinham : :py:class:`.SpinHamiltonian`
        Spin Hamiltonian
    distance_digits : int, default 5
        Precision for comparing two linear distances.
    plot_dmi : bool, default True
        Whether to plot DMI vectors
    dmi_scale : float, default 1.0
        Scale for the maximum dmi vector length.

    Returns
    -------
    pe1 : :py:class:`.PlotlyEngine`
        Instance of the magnopy's plot engine, with plotted spin Hamiltonian. Ready to be
        saved or showed. Only on-site parameters are plotted
    pe2 : :py:class:`.PlotlyEngine`
        Instance of the magnopy's plot engine, with plotted spin Hamiltonian. Ready to be
        saved or showed. Only two-spins/two-sites parameters are plotted.
    """

    pe1 = PlotlyEngine(_sphinx_gallery_fix=_sphinx_gallery_fix)

    pe1.plot_cell(
        cell=spinham.cell,
        color="LightGrey",
        legend_label="(0, 0, 0) unit cell",
        legend_group="(0, 0, 0) unit cell",
    )

    points = spinham.magnetic_atoms.positions @ spinham.cell

    hoverinfo = [
        "<br>".join(
            [
                f'Magnetic center "{name}", located at',
                f"  Relative: ({pos[0]:.8f}, {pos[1]:.8f}, {pos[2]:.8f})",
                f"  Absolute: ({(pos @ spinham.cell)[0]:.8f}, {(pos @ spinham.cell)[1]:.8f}, {(pos @ spinham.cell)[2]:.8f})",
                "",
                f"With spin value {spin:.3f}",
                "",
                f"and g-factor {g_factor:.3f}",
            ]
        )
        for name, pos, spin, g_factor in zip(
            spinham.magnetic_atoms.names,
            spinham.magnetic_atoms.positions,
            spinham.magnetic_atoms.spins,
            spinham.magnetic_atoms.g_factors,
        )
    ]

    for alpha, parameter in spinham.p1:
        alpha = spinham.map_to_magnetic[alpha]

        hoverinfo[alpha] += "<br>".join(
            [
                "",
                "One-spin parameter:",
                f"  {parameter[0]:10.5f} {parameter[0]:10.5f} {parameter[0]:10.5f}",
                "",
            ]
        )

    for alpha, parameter in spinham.p21:
        alpha = spinham.map_to_magnetic[alpha]

        hoverinfo[alpha] += "<br>".join(
            [
                "",
                "Two-spin parameter:",
                f"  {parameter[0][0]:10.5f} {parameter[1][0]:10.5f} {parameter[2][0]:10.5f}",
                f"  {parameter[0][1]:10.5f} {parameter[1][1]:10.5f} {parameter[2][1]:10.5f}",
                f"  {parameter[0][2]:10.5f} {parameter[1][2]:10.5f} {parameter[2][2]:10.5f}",
                "",
            ]
        )

    with_parameters = []
    without_parameters = []

    for i, text in enumerate(hoverinfo):
        if text == "":
            without_parameters.append(i)
        else:
            with_parameters.append(i)

    xyz = points[with_parameters].T
    pe1.fig.add_traces(
        data=dict(
            type="scatter3d",
            mode="markers",
            legendgroup="with parameters",
            name="Magnetic cites with parameters\n(hover to see the values)",
            showlegend=True,
            x=xyz[0],
            y=xyz[1],
            z=xyz[2],
            marker=dict(size=10, color="Black"),
            text=[hoverinfo[i] for i in with_parameters],
            hoverinfo="text",
        ),
    )

    if len(without_parameters) != 0:
        xyz = points[without_parameters].T
        pe1.fig.add_traces(
            data=dict(
                type="scatter3d",
                mode="markers",
                legendgroup="without parameters",
                name="Magnetic cites without parameters",
                showlegend=True,
                x=xyz[0],
                y=xyz[1],
                z=xyz[2],
                marker=dict(size=5, color="Grey"),
                hoverinfo="none",
            ),
        )

    # Fix for plotly #7143
    points = points.T
    pe1._update_range(
        x_min=points[0].min(),
        x_max=points[0].max(),
        y_min=points[1].min(),
        y_max=points[1].max(),
        z_min=points[2].min(),
        z_max=points[2].max(),
    )

    pe2 = PlotlyEngine(_sphinx_gallery_fix=_sphinx_gallery_fix)

    start_points = []
    end_points = []
    parameters = []
    distances = []
    dmi_vectors = []
    dmi_positions = []
    alphas = []
    betas = []
    nus = []

    cells_to_plot = []
    for alpha, beta, nu, parameter in spinham.p22:
        vector = get_vector(
            cell=spinham.cell, atoms=spinham.atoms, atom1=alpha, atom2=beta, R=nu
        )
        start_points.append(spinham.atoms.positions[alpha] @ spinham.cell)
        end_points.append(start_points[-1] + vector)
        distances.append(round(np.linalg.norm(vector), ndigits=distance_digits))
        parameters.append(parameter)
        alphas.append(alpha)
        betas.append(beta)
        nus.append(nu)

        dmi_vectors.append(to_dmi(parameter=parameter))
        dmi_positions.append(start_points[-1] + vector / 2)

        if nu != (0, 0, 0):
            cells_to_plot.append(nu)

    indices = np.argsort(distances)

    start_points = np.array(start_points)[indices]
    end_points = np.array(end_points)[indices]
    parameters = np.array(parameters)[indices]
    distances = np.array(distances)[indices]
    dmi_positions = np.array(dmi_positions)[indices]
    dmi_vectors = np.array(dmi_vectors)[indices]
    alphas = np.array(alphas, dtype=int)[indices]
    betas = np.array(betas, dtype=int)[indices]
    nus = np.array(nus, dtype=int)[indices]

    dmi_max_length = np.array([np.linalg.norm(_) for _ in dmi_vectors]).max()

    if dmi_max_length != 0:
        dmi_vectors = dmi_vectors / dmi_max_length * dmi_vectors_scale

    unique_distances = np.unique(distances)

    chars = "0123456789ABCDEF"
    random_colors = [
        "#" + "".join(random.sample(chars, 6)) for _ in range(len(unique_distances))
    ]
    colors = dict(zip(unique_distances, random_colors))

    colors = [colors[_] for _ in distances]

    legend_labels = [f"d = {_:.{distance_digits}f}" for _ in unique_distances]

    legend_groups = dict(zip(unique_distances, legend_labels))
    legend_labels = dict(zip(unique_distances, legend_labels))

    hoverinfo = [
        "<br>".join(
            [
                "Bond",
                f'  from magnetic center "{spinham.atoms.names[alpha]}" in (0, 0, 0) unit cell',
                f'  to magnetic center "{spinham.atoms.names[beta]}" in ({nu[0]}, {nu[1]}, {nu[2]}) unit cell',
                "",
                "Full two-spins/two-sites parameter:",
                f"  {param[0][0]:10.5f} {param[0][1]:10.5f} {param[0][2]:10.5f}",
                f"  {param[1][0]:10.5f} {param[1][1]:10.5f} {param[1][2]:10.5f}",
                f"  {param[2][0]:10.5f} {param[2][1]:10.5f} {param[2][2]:10.5f}",
                "",
                f"Isotropic: {to_iso(param):10.5f}",
                "",
                f"DMI: {to_dmi(param)[0]:10.5f} {to_dmi(param)[1]:10.5f} {to_dmi(param)[2]:10.5f}",
            ]
        )
        for alpha, beta, nu, param in zip(alphas, betas, nus, parameters)
    ]

    legend_label = "Other_cells"
    magnetic_sites = []
    for i, j, k in cells_to_plot:
        shift = [i, j, k] @ spinham.cell
        pe2.plot_cell(
            cell=spinham.cell,
            legend_group="other cells",
            legend_label=legend_label,
            color="LightGrey",
            shift=shift,
            plot_vectors=False,
        )
        if legend_label is not None:
            legend_label = None

        magnetic_sites.extend(
            spinham.magnetic_atoms.positions @ spinham.cell + shift[np.newaxis, :]
        )
    pe2.plot_cell(
        cell=spinham.cell,
        legend_label="(0, 0, 0) unit cell",
        color="Grey",
        plot_vectors=True,
    )

    pe2.plot_points(
        points=magnetic_sites,
        colors="LightGrey",
        legend_group="Magnetic sites",
        scale=2,
    )
    pe2.plot_points(
        points=spinham.magnetic_atoms.positions @ spinham.cell,
        colors="Grey",
        legend_label="Magnetic sites",
        legend_group="Magnetic sites",
        scale=2,
    )

    for i in range(len(parameters)):
        x, y, z = np.array([start_points[i], end_points[i]]).T
        pe2.fig.add_traces(
            data=dict(
                type="scatter3d",
                mode="lines",
                legendgroup=legend_groups[distances[i]],
                name=legend_labels[distances[i]],
                showlegend=legend_labels[distances[i]] is not None,
                x=x,
                y=y,
                z=z,
                line=dict(color=colors[i], width=4),
                hoverinfo="text",
                text=hoverinfo[i],
            ),
        )
        if plot_dmi:
            pe2.plot_vector(
                start_point=dmi_positions[i] - dmi_vectors[i] / 2,
                end_point=dmi_positions[i] + dmi_vectors[i] / 2,
                color=colors[i],
                legend_group=legend_groups[distances[i]],
            )

        if legend_labels[distances[i]] is not None:
            legend_labels[distances[i]] = None

    return pe1, pe2


def change_cell(spinham, new_cell, new_atoms_specs):
    r"""
    Changes the unit cell of a spin Hamiltonian.

    .. warning::
        Experimental feature. Not tested well.

    Parameters
    ----------
    spinham : :py:class:`.SpinHamiltonian`
    new_cell : (3, 3) |array-like|_
    new_atoms_specs : list of tuple

        .. code-block:: python

            new_atoms_specs = [(index, nu), ...]

        where ``index`` is an index of an atom in ``spinham.atoms`` and ``nu = (i,j,k)``
        is the unit cell to which an atom belongs.

    Returns
    -------
    new_spinham : :py:class:`.SpinHamiltonian`
    """

    new_cell = np.array(new_cell)

    new_atoms = {}

    for key in spinham.atoms:
        new_atoms[key] = []

    for atom_index, (i, j, k) in new_atoms_specs:
        for key in spinham.atoms:
            if key == "positions":
                position = spinham.atoms.positions[atom_index]
                new_position = (
                    np.array([(position[0] + i), (position[1] + j), (position[2] + k)])
                    @ spinham.cell
                    @ np.linalg.inv(new_cell)
                )
                new_atoms["positions"].append(new_position)
            else:
                new_atoms[key].append(spinham.atoms[key][atom_index])

    map_to_indices = {
        tuple(np.around(new_atoms["positions"][i], decimals=2)): i
        for i in range(len(new_atoms["positions"]))
    }

    new_spinham = SpinHamiltonian(
        cell=new_cell, atoms=new_atoms, convention=spinham.convention
    )

    def get_new_indices(atom_index, ijk):
        i, j, k = ijk
        position = spinham.atoms.positions[atom_index]
        new_position = (
            np.array([(position[0] + i), (position[1] + j), (position[2] + k)])
            @ spinham.cell
            @ np.linalg.inv(new_cell)
        )

        new_ijk = new_position // 1

        new_ijk = tuple(map(int, new_ijk))

        new_index = map_to_indices[tuple(np.around(new_position % 1, decimals=2))]

        return new_index, new_ijk

    # Re-populate parameters
    for new_alpha, (atom_index, (i, j, k)) in enumerate(new_atoms_specs):
        # One spin
        for alpha, parameter in spinham._1:
            if alpha == atom_index:
                new_spinham.add_1(alpha=new_alpha, parameter=parameter)

        # Two spins
        for alpha, parameter in spinham._21:
            if alpha == atom_index:
                new_spinham.add_21(alpha=new_alpha, parameter=parameter)

        for alpha, beta, nu, parameter in spinham._22:
            if alpha == atom_index:
                new_beta, new_nu = get_new_indices(
                    atom_index=beta, ijk=np.array(nu) + (i, j, k)
                )

                new_spinham.add_22(
                    alpha=new_alpha,
                    beta=new_beta,
                    nu=new_nu,
                    parameter=parameter,
                    when_present="replace",
                )

        # Three spins
        for alpha, parameter in spinham._31:
            if alpha == atom_index:
                new_spinham.add_31(alpha=new_alpha, parameter=parameter)

        for alpha, beta, nu, parameter in spinham._32:
            if alpha == atom_index:
                new_beta, new_nu = get_new_indices(
                    atom_index=beta, ijk=np.array(nu) + (i, j, k)
                )

                new_spinham.add_32(
                    alpha=new_alpha,
                    beta=new_beta,
                    nu=new_nu,
                    parameter=parameter,
                    when_present="replace",
                )

        for alpha, beta, gamma, nu, _lambda, parameter in spinham._33:
            if alpha == atom_index:
                new_beta, new_nu = get_new_indices(
                    atom_index=beta, ijk=np.array(nu) + (i, j, k)
                )
                new_gamma, new_lambda = get_new_indices(
                    atom_index=gamma, ijk=np.array(_lambda) + (i, j, k)
                )

                new_spinham.add_33(
                    alpha=new_alpha,
                    beta=new_beta,
                    gamma=new_gamma,
                    nu=new_nu,
                    _lambda=new_lambda,
                    parameter=parameter,
                    when_present="replace",
                )

        # Four spins
        for alpha, parameter in spinham._41:
            if alpha == atom_index:
                new_spinham.add_41(alpha=new_alpha, parameter=parameter)

        for alpha, beta, nu, parameter in spinham._421:
            if alpha == atom_index:
                new_beta, new_nu = get_new_indices(
                    atom_index=beta, ijk=np.array(nu) + (i, j, k)
                )

                new_spinham.add_421(
                    alpha=new_alpha,
                    beta=new_beta,
                    nu=new_nu,
                    parameter=parameter,
                    when_present="replace",
                )

        for alpha, beta, nu, parameter in spinham._422:
            if alpha == atom_index:
                new_beta, new_nu = get_new_indices(
                    atom_index=beta, ijk=np.array(nu) + (i, j, k)
                )

                new_spinham.add_422(
                    alpha=new_alpha,
                    beta=new_beta,
                    nu=new_nu,
                    parameter=parameter,
                    when_present="replace",
                )

        for alpha, beta, gamma, nu, _lambda, parameter in spinham._43:
            new_beta, new_nu = get_new_indices(
                atom_index=beta, ijk=np.array(nu) + (i, j, k)
            )
            new_gamma, new_lambda = get_new_indices(
                atom_index=gamma, ijk=np.array(_lambda) + (i, j, k)
            )

            new_spinham.add_43(
                alpha=new_alpha,
                beta=new_beta,
                gamma=new_gamma,
                nu=new_nu,
                _lambda=new_lambda,
                parameter=parameter,
                when_present="replace",
            )

        for (
            alpha,
            beta,
            gamma,
            epsilon,
            nu,
            _lambda,
            rho,
            parameter,
        ) in spinham._44:
            new_beta, new_nu = get_new_indices(
                atom_index=beta, ijk=np.array(nu) + (i, j, k)
            )
            new_gamma, new_lambda = get_new_indices(
                atom_index=gamma, ijk=np.array(_lambda) + (i, j, k)
            )
            new_epsilon, new_rho = get_new_indices(
                atom_index=epsilon, ijk=np.array(rho) + (i, j, k)
            )

            new_spinham.add_44(
                alpha=new_alpha,
                beta=new_beta,
                gamma=new_gamma,
                epsilon=new_epsilon,
                nu=new_nu,
                _lambda=new_lambda,
                rho=new_rho,
                parameter=parameter,
                when_present="replace",
            )

    return new_spinham


# DEPRECATED in v0.4.0
# Remove in May 2026
def plot_dispersion(data, kp=None, output_filename=None, ylabel=None):
    r"""
    Plots some k-resolved data.

    .. deprecated:: 0.4.0
        Function deprecated in 0.4.0 and will be removed in May 2026. Use :py:func:`magnopy.io.plot_dispersion()` instead.

    If only the ``data`` are given, then an index of the omegas is used for abscissa (x
    axis).

    Parameters
    ----------
    data : (N, M) |array-like|_
        Some k-resolved data. N (:math:`\ge 1`) is the amount of kpoints. M is the
        number of data modes/entries.  Expected to be given in meV.
    kp : :py:class:`wulfric.Kpoints`, optional.
        Instance of the :py:class:`wulfric.Kpoints` class. It should be the same
        instance that were used in the preparation of the ``data``.
    output_filename : str, optional
        Name of the file for saving the image. If ``None``, then the graph would be
        opened in the interactive matplotlib window.
    ylabel : str, optional
        Label for the ordinate (y axis). Do not include units, units are included automatically.
    """

    import warnings

    warnings.warn(
        "Function magnopy.experimental.plot_dispersion() was deprecated in 0.4.0 and will be removed in May 2026. Use magnopy.io.plot_dispersion() instead",
        DeprecationWarning,
        stacklevel=2,
    )

    if not MATPLOTLIB_AVAILABLE:
        import warnings

        warnings.warn(MATPLOTLIB_ERROR_MESSAGE, RuntimeWarning, stacklevel=2)

        return

    data = np.array(data).T

    fig, ax = plt.subplots()

    if len(data.shape) == 2:
        for entry in data:
            if kp is not None:
                ax.plot(kp.flat_points(), entry, lw=1, color="#A47864")
            else:
                ax.plot(entry, lw=1, color="#A47864")
                ax.set_xlim(0, len(entry))
    else:
        if kp is not None:
            ax.plot(kp.flat_points(), data, lw=1, color="#A47864")
        else:
            ax.plot(data, lw=1, color="#A47864")
            ax.set_xlim(0, len(data))

    if kp is not None:
        ax.set_xticks(kp.ticks(), kp.labels, fontsize=13)
        ax.set_xlim(kp.ticks()[0], kp.ticks()[-1])
        ax.vlines(
            kp.ticks(),
            0,
            1,
            lw=0.5,
            color="grey",
            ls="dashed",
            zorder=0,
            transform=ax.get_xaxis_transform(),
        )

    if ylabel is not None:
        ax.set_ylabel(f"{ylabel}, meV", fontsize=15)

    ax.hlines(
        0,
        0,
        1,
        lw=0.5,
        color="grey",
        linestyle="dashed",
        transform=ax.get_yaxis_transform(),
    )

    # Add twin axis

    ylims = ax.get_ylim()

    meV_to_THz = _MAGNON_ENERGY_UNITS["mev"] / _MAGNON_ENERGY_UNITS["thz"]

    twinax = ax.twinx()

    twinax.set_ylim(meV_to_THz * ylims[0], meV_to_THz * ylims[1])

    twinax.set_ylabel(f"{ylabel}, THz", fontsize=15)

    if output_filename is not None:
        plt.savefig(output_filename, dpi=400, bbox_inches="tight")
    else:
        plt.show()

    plt.close()



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================


from copy import deepcopy
from math import ceil

import numpy as np
from wulfric import add_sugar

from magnopy._spinham._c1 import _add_1, _p1, _remove_1
from magnopy._spinham._c21 import _add_21, _p21, _remove_21
from magnopy._spinham._c22 import _add_22, _get_primary_p22, _p22, _remove_22
from magnopy._spinham._c31 import _add_31, _p31, _remove_31
from magnopy._spinham._c32 import _add_32, _p32, _remove_32
from magnopy._spinham._c33 import _add_33, _p33, _remove_33
from magnopy._spinham._c41 import _add_41, _p41, _remove_41
from magnopy._spinham._c43 import _add_43, _p43, _remove_43
from magnopy._spinham._c44 import _add_44, _p44, _remove_44
from magnopy._spinham._c421 import _add_421, _p421, _remove_421
from magnopy._spinham._c422 import _add_422, _p422, _remove_422
from magnopy._spinham._convention import Convention

from magnopy._data_validation import _validated_units

from magnopy._constants._units import _PARAMETER_UNITS, _PARAMETER_UNITS_MAKEUP
from magnopy._constants._si import BOHR_MAGNETON, ANGSTROM, VACUUM_MAGNETIC_PERMEABILITY

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def _merge(list1: list, list2: list) -> list:
    r"""
    Merge two sorted parameter lists for any term.

    Lists of parameters have the form

    .. code-block:: python

        list = [[specs, parameter], ...]

    Comparison is based on specs.

    Parameter
    ---------
    list1 : list
        First list of parameters.
    list2 : list
        Second list of parameters.

    Returns
    -------
    merged_list : list
        Merged list of parameters.
    """

    list1 = deepcopy(list1)
    list2 = deepcopy(list2)

    merged_list = []

    i1 = 0
    i2 = 0

    while i1 < len(list1) or i2 < len(list2):
        if i1 >= len(list1):
            merged_list.append(list2[i2])
            i2 += 1
        elif i2 >= len(list2):
            merged_list.append(list1[i1])
            i1 += 1
        elif list1[i1][:-1] == list2[i2][:-1]:
            merged_list.append(list1[i1])
            merged_list[-1][-1] = merged_list[-1][-1] + list2[i2][-1]
            i1 += 1
            i2 += 1
        elif list1[i1][:-1] < list2[i2][:-1]:
            merged_list.append(list1[i1])
            i1 += 1
        else:
            merged_list.append(list2[i2])
            i2 += 1

    return merged_list


class SpinHamiltonian:
    r"""
    Spin Hamiltonian.

    Parameters
    ----------

    convention : :py:class:`.Convention` or str
        A convention of the spin Hamiltonian.

    cell : (3, 3) |array-like|_
        Matrix of a cell, rows are interpreted as vectors.

    atoms : dict
        Dictionary with atoms.

    units : str, default "meV"
        .. versionadded:: 0.3.0

        Units of the Hamiltonian's parameters. See :py:attr:`.SpinHamiltonian.units`
        for more details. Case-insensitive.


    Examples
    --------

    For example of usage see the page in the user guide -
    :ref:`user-guide_usage_spin-hamiltonian`.

    """

    def __init__(self, cell, atoms, convention, units="meV") -> None:
        self._cell = np.array(cell)

        self._atoms = add_sugar(atoms)

        # Only the magnetic sites
        self._magnetic_atoms = None
        self._map_to_magnetic = None
        self._map_to_all = None

        self._convention = convention

        if units.lower() not in _PARAMETER_UNITS:
            raise ValueError(
                f'Given units ("{units}") are not supported. Please use one of\n  * '
                + "\n  * ".join(list(_PARAMETER_UNITS))
            )

        self._units = units.lower()

        # [[alpha, parameter], ...]
        self._1 = []

        # [[alpha, parameter], ...]
        self._21 = []

        # [[alpha, beta, nu, parameter], ...]
        self._22 = []

        # [[alpha, parameter], ...]
        self._31 = []

        # [[alpha, beta, nu, parameter], ...]
        self._32 = []

        # [[alpha, beta, gamma, nu, lambda, parameter], ...]
        self._33 = []

        # [[alpha, parameter], ...]
        self._41 = []

        # [[alpha, beta, nu, parameter], ...]
        self._421 = []

        # [[alpha, beta, nu, parameter], ...]
        self._422 = []

        # [[alpha, beta, gamma, nu, lambda, parameter], ...]
        self._43 = []

        # [[alpha, beta, gamma, epsilon, nu, lambda, rho, parameter], ...]
        self._44 = []

    ############################################################################
    #                              Cell and Atoms                              #
    ############################################################################

    @property
    def cell(self):
        r"""
        Cell of the crystal on which the Hamiltonian is build.

        Returns
        -------

        cell : (3, 3) :numpy:`ndarray`
            Matrix of the cell, rows are lattice vectors.

        Notes
        -----

        If is not recommended to change the ``cell`` property after the creation of
        :py:class:`.SpinHamiltonian`. In fact an attempt to do so will raise an
        ``AttributeError``:

        .. doctest::

            >>> import numpy as np
            >>> import magnopy
            >>> convention = magnopy.Convention()
            >>> spinham = magnopy.SpinHamiltonian(
            ...     cell=np.eye(3), atoms={}, convention=convention
            ... )
            >>> spinham.cell = 2 * np.eye(3)
            Traceback (most recent call last):
            ...
            AttributeError: Change of the cell attribute is not supported after the creation of SpinHamiltonian instance. If you need to modify cell, then use pre-defined methods of SpinHamiltonian or create a new one.

        Use pre-defined methods of the :py:class:`.SpinHamiltonian` class to safely
        modify the cell.

        If you need to change the cell attribute, then use

        .. doctest::

            >>> import numpy as np
            >>> import magnopy
            >>> convention = magnopy.Convention()
            >>> spinham = magnopy.SpinHamiltonian(
            ...     cell=np.eye(3), atoms={}, convention=convention
            ... )
            >>> spinham.cell
            array([[1., 0., 0.],
                   [0., 1., 0.],
                   [0., 0., 1.]])
            >>> spinham._cell = 2 * np.eye(3)
            >>> spinham.cell
            array([[2., 0., 0.],
                   [0., 2., 0.],
                   [0., 0., 2.]])


        In the latter case correct behavior of magnopy **is not** guaranteed. Use only
        if you have a deep understanding of the magnopy source code.
        """

        return self._cell

    @cell.setter
    def cell(self, new_value):
        raise AttributeError(
            "Change of the cell attribute is not allowed after the creation of SpinHamiltonian instance. SpinHamiltonian.cell is immutable."
        )

    @property
    def atoms(self) -> dict:
        r"""
        Atoms of the crystal on which the Hamiltonian is build.

        Returns
        -------

        atoms : dict (with added sugar)
            Dictionary with the atoms.

        Notes
        -----

        If is not recommended to change the atoms property after the creation of
        :py:class:`.SpinHamiltonian`. In fact an attempt to do so will raise an
        ``AttributeError``:

        .. doctest::

            >>> import numpy as np
            >>> import magnopy
            >>> convention = magnopy.Convention()
            >>> spinham = magnopy.SpinHamiltonian(
            ...     cell=np.eye(3), atoms={}, convention=convention
            ... )
            >>> spinham.atoms = {"names": ["Cr"]}
            Traceback (most recent call last):
            ...
            AttributeError: Change of the atoms dictionary is not supported after the creation of SpinHamiltonian instance. If you need to modify atoms, then use pre-defined methods of SpinHamiltonian or create a new one.

        Use pre-defined methods of the :py:class:`.SpinHamiltonian` class to safely
        modify atoms.

        If you need to change the whole dictionary at once, then use

        .. doctest::

            >>> import numpy as np
            >>> import magnopy
            >>> convention = magnopy.Convention()
            >>> spinham = magnopy.SpinHamiltonian(
            ...     cell=np.eye(3), atoms={}, convention=convention
            ... )
            >>> spinham.atoms
            {}
            >>> spinham._atoms = {"names": ["Cr"]}
            >>> spinham.atoms
            {'names': ['Cr']}

        In the latter case correct behavior of magnopy **is not** guaranteed. Use only
        if you have a deep understanding of the magnopy source code.
        """

        return self._atoms

    @atoms.setter
    def atoms(self, new_value):
        raise AttributeError(
            "Change of the atoms dictionary is not supported after the creation of SpinHamiltonian instance. SpinHamiltonian.atoms is immutable."
        )

    def _reset_internals(self):
        self._map_to_magnetic = None
        self._map_to_all = None
        self._magnetic_atoms = None

    def _update_internals(self):
        # Identify magnetic sites
        indices = set()

        for alpha, _ in self._1:
            indices.add(alpha)

        for alpha, _ in self._21:
            indices.add(alpha)

        for alpha, beta, _, _ in self._22:
            indices.add(alpha)
            indices.add(beta)

        for alpha, _ in self._31:
            indices.add(alpha)

        for alpha, beta, _, _ in self._32:
            indices.add(alpha)
            indices.add(beta)

        for alpha, beta, gamma, _, _, _ in self._33:
            indices.add(alpha)
            indices.add(beta)
            indices.add(gamma)

        for alpha, _ in self._41:
            indices.add(alpha)

        for alpha, beta, _, _ in self._421:
            indices.add(alpha)
            indices.add(beta)

        for alpha, beta, _, _ in self._422:
            indices.add(alpha)
            indices.add(beta)

        for alpha, beta, gamma, _, _, _ in self._43:
            indices.add(alpha)
            indices.add(beta)
            indices.add(gamma)

        for alpha, beta, gamma, epsilon, _, _, _, _ in self._44:
            indices.add(alpha)
            indices.add(beta)
            indices.add(gamma)
            indices.add(epsilon)

        indices = sorted(list(indices))

        # Create index map from all to magnetic
        self._map_to_magnetic = [None for _ in range(len(self.atoms.names))]
        for i in range(len(indices)):
            self._map_to_magnetic[indices[i]] = i

        # Create index map from magnetic to all
        self._map_to_all = indices

        # Create magnetic atoms dictionary
        self._magnetic_atoms = add_sugar({})
        for key in self.atoms:
            self._magnetic_atoms[key] = []

            for full_index in indices:
                self._magnetic_atoms[key].append(self.atoms[key][full_index])

    @property
    def map_to_magnetic(self):
        r"""
        Index map from all atoms to the magnetic ones.

        Returns
        -------

        map_to_magnetic (L, ) list of int
            Index map. Integers. ``L = len(spinham.atoms.names)``
        """

        if self._map_to_magnetic is None:
            self._update_internals()

        return self._map_to_magnetic

    @property
    def map_to_all(self):
        r"""
        Index map from magnetic atoms to all atoms.

        Returns
        -------

        map_to_all (M, ) list of int
            Index map. Integers. ``M = len(spinham.magnetic_atoms.names)``
        """

        if self._map_to_all is None:
            self._update_internals()

        return self._map_to_all

    @property
    def magnetic_atoms(self):
        r"""
        Magnetic atoms of the spin Hamiltonian.

        Magnetic atom is defined as an atom with at least one parameter associated with
        it.

        This property is dynamically computed at every call.

        Returns
        -------

        magnetic_atoms : list of int
            Indices of magnetic atoms in the ``spinham.atoms``. Sorted.

        See Also
        --------

        M
        """

        if self._magnetic_atoms is None:
            self._update_internals()

        return self._magnetic_atoms

    @property
    def M(self):
        r"""
        Number of spins (magnetic atoms) in the unit cell.

        Returns
        -------

        M : int
            Number of spins (magnetic atoms) in the unit cell.

        See Also
        --------

        magnetic_atoms
        """

        return len(self.magnetic_atoms.names)

    ############################################################################
    #                                Convention                                #
    ############################################################################
    @property
    def convention(self) -> Convention:
        r"""
        Convention of the spin Hamiltonian.

        Returns
        -------

        convention : :py:class:`.Convention`

        See Also
        --------

        Convention
        """

        return self._convention

    @convention.setter
    def convention(self, new_convention: Convention):
        self._set_multiple_counting(new_convention._multiple_counting)

        self._set_spin_normalization(new_convention._spin_normalized)

        self._set_c1(new_convention._c1)

        self._set_c21(new_convention._c21)
        self._set_c22(new_convention._c22)

        self._set_c31(new_convention._c31)
        self._set_c32(new_convention._c32)
        self._set_c33(new_convention._c33)

        self._set_c41(new_convention._c41)
        self._set_c421(new_convention._c421)
        self._set_c422(new_convention._c422)
        self._set_c43(new_convention._c43)
        self._set_c44(new_convention._c44)

        self._convention = new_convention

    def _set_multiple_counting(self, multiple_counting: bool) -> None:
        if multiple_counting is None or self.convention._multiple_counting is None:
            return

        multiple_counting = bool(multiple_counting)

        if self.convention.multiple_counting == multiple_counting:
            return

        # It was absent before
        if multiple_counting:
            factor = 0.5
        # It was present before
        else:
            factor = 2.0

        # For (two spins & two sites)
        for index in range(len(self._22)):
            self._22[index][3] = self._22[index][3] * factor

        # For (three spins & two sites)
        for index in range(len(self._32)):
            self._32[index][3] = self._32[index][3] * factor

        # For (four spins & two sites (3+1))
        for index in range(len(self._421)):
            self._421[index][3] = self._421[index][3] * factor

        # For (four spins & two sites (2+2))
        for index in range(len(self._422)):
            self._422[index][3] = self._422[index][3] * factor

        # It was absent before
        if multiple_counting:
            factor = 1 / 6
        # It was present before
        else:
            factor = 6

        # For (three spins & three sites)
        for index in range(len(self._33)):
            self._33[index][5] = self._33[index][5] * factor

        # For (four spins & three sites)
        for index in range(len(self._43)):
            self._43[index][5] = self._43[index][5] * factor

        # It was absent before
        if multiple_counting:
            factor = 1 / 24
        # It was present before
        else:
            factor = 24

        # For (four spins & four sites)
        for index in range(len(self._44)):
            self._44[index][7] = self._44[index][7] * factor

    def _set_spin_normalization(self, spin_normalized: bool) -> None:
        if spin_normalized is None or self.convention._spin_normalized is None:
            return

        spin_normalized = bool(spin_normalized)

        if self.convention.spin_normalized == spin_normalized:
            return

        # Before it was not normalized
        if spin_normalized:
            # For (one spin & one site)
            for index in range(len(self._1)):
                alpha = self._1[index][0]
                self._1[index][1] = self._1[index][1] * self.atoms.spins[alpha]
            # For (two spins & one site)
            for index in range(len(self._21)):
                alpha = self._21[index][0]
                self._21[index][1] = self._21[index][1] * self.atoms.spins[alpha] ** 2
            # For (two spins & two sites)
            for index in range(len(self._22)):
                alpha = self._22[index][0]
                beta = self._22[index][1]
                self._22[index][3] = self._22[index][3] * (
                    self.atoms.spins[alpha] * self.atoms.spins[beta]
                )
            # For (three spins & one site)
            for index in range(len(self._31)):
                alpha = self._31[index][0]
                self._31[index][1] = self._31[index][1] * self.atoms.spins[alpha] ** 3
            # For (three spins & two sites)
            for index in range(len(self._32)):
                alpha = self._32[index][0]
                beta = self._32[index][1]
                self._32[index][3] = self._32[index][3] * (
                    self.atoms.spins[alpha] ** 2 * self.atoms.spins[beta]
                )
            # For (three spins & three sites)
            for index in range(len(self._33)):
                alpha = self._33[index][0]
                beta = self._33[index][1]
                gamma = self._33[index][2]
                self._33[index][5] = self._33[index][5] * (
                    self.atoms.spins[alpha]
                    * self.atoms.spins[beta]
                    * self.atoms.spins[gamma]
                )
            # For (four spins & one site)
            for index in range(len(self._41)):
                alpha = self._41[index][0]
                self._41[index][1] = self._41[index][1] * self.atoms.spins[alpha] ** 4
            # For (four spins & two sites (3+1))
            for index in range(len(self._421)):
                alpha = self._421[index][0]
                beta = self._421[index][1]
                self._421[index][3] = self._421[index][3] * (
                    self.atoms.spins[alpha] ** 3 * self.atoms.spins[beta]
                )
            # For (four spins & two sites (2+2))
            for index in range(len(self._422)):
                alpha = self._422[index][0]
                beta = self._422[index][1]
                self._422[index][3] = self._422[index][3] * (
                    self.atoms.spins[alpha] ** 2 * self.atoms.spins[beta] ** 2
                )
            # For (four spins & three sites)
            for index in range(len(self._43)):
                alpha = self._43[index][0]
                beta = self._43[index][1]
                gamma = self._43[index][2]
                self._43[index][5] = self._43[index][5] * (
                    self.atoms.spins[alpha] ** 2
                    * self.atoms.spins[beta]
                    * self.atoms.spins[gamma]
                )
            # For (four spins & four sites)
            for index in range(len(self._44)):
                alpha = self._44[index][0]
                beta = self._44[index][1]
                gamma = self._44[index][2]
                epsilon = self._44[index][3]
                self._44[index][7] = self._44[index][7] * (
                    self.atoms.spins[alpha]
                    * self.atoms.spins[beta]
                    * self.atoms.spins[gamma]
                    * self.atoms.spins[epsilon]
                )
        # Before it was normalized
        else:
            # For (one spin & one site)
            for index in range(len(self._1)):
                alpha = self._1[index][0]
                self._1[index][1] = self._1[index][1] / self.atoms.spins[alpha]
            # For (two spins & one site)
            for index in range(len(self._21)):
                alpha = self._21[index][0]
                self._21[index][1] = self._21[index][1] / self.atoms.spins[alpha] ** 2
            # For (two spins & two sites)
            for index in range(len(self._22)):
                alpha = self._22[index][0]
                beta = self._22[index][1]
                self._22[index][3] = self._22[index][3] / (
                    self.atoms.spins[alpha] * self.atoms.spins[beta]
                )
            # For (three spins & one site)
            for index in range(len(self._31)):
                alpha = self._31[index][0]
                self._31[index][1] = self._31[index][1] / self.atoms.spins[alpha] ** 3
            # For (three spins & two sites)
            for index in range(len(self._32)):
                alpha = self._32[index][0]
                beta = self._32[index][1]
                self._32[index][3] = self._32[index][3] / (
                    self.atoms.spins[alpha] ** 2 * self.atoms.spins[beta]
                )
            # For (three spins & three sites)
            for index in range(len(self._33)):
                alpha = self._33[index][0]
                beta = self._33[index][1]
                gamma = self._33[index][2]
                self._33[index][5] = self._33[index][5] / (
                    self.atoms.spins[alpha]
                    * self.atoms.spins[beta]
                    * self.atoms.spins[gamma]
                )
            # For (four spins & one site)
            for index in range(len(self._41)):
                alpha = self._41[index][0]
                self._41[index][1] = self._41[index][1] / self.atoms.spins[alpha] ** 4
            # For (four spins & two sites (3+1))
            for index in range(len(self._421)):
                alpha = self._421[index][0]
                beta = self._421[index][1]
                self._421[index][3] = self._421[index][3] / (
                    self.atoms.spins[alpha] ** 3 * self.atoms.spins[beta]
                )
            # For (four spins & two sites (2+2))
            for index in range(len(self._422)):
                alpha = self._422[index][0]
                beta = self._422[index][1]
                self._422[index][3] = self._422[index][3] / (
                    self.atoms.spins[alpha] ** 2 * self.atoms.spins[beta] ** 2
                )
            # For (four spins & three sites)
            for index in range(len(self._43)):
                alpha = self._43[index][0]
                beta = self._43[index][1]
                gamma = self._43[index][2]
                self._43[index][5] = self._43[index][5] / (
                    self.atoms.spins[alpha] ** 2
                    * self.atoms.spins[beta]
                    * self.atoms.spins[gamma]
                )
            # For (four spins & four sites)
            for index in range(len(self._44)):
                alpha = self._44[index][0]
                beta = self._44[index][1]
                gamma = self._44[index][2]
                epsilon = self._44[index][3]
                self._44[index][7] = self._44[index][7] / (
                    self.atoms.spins[alpha]
                    * self.atoms.spins[beta]
                    * self.atoms.spins[gamma]
                    * self.atoms.spins[epsilon]
                )

    def _set_c1(self, new_c1: float) -> None:
        if new_c1 is None or self.convention._c1 is None:
            return

        new_c1 = float(new_c1)

        if self.convention.c1 == new_c1:
            return

        # If factor is changing one has to scale parameters.
        for index in range(len(self._1)):
            self._1[index][1] = self._1[index][1] * self.convention.c1 / new_c1

    def _set_c21(self, new_c21: float) -> None:
        if new_c21 is None or self.convention._c21 is None:
            return

        new_c21 = float(new_c21)

        if self.convention.c21 == new_c21:
            return

        # If factor is changing one has to scale parameters.
        for index in range(len(self._21)):
            self._21[index][1] = self._21[index][1] * self.convention.c21 / new_c21

    def _set_c22(self, new_c22: float) -> None:
        if new_c22 is None or self.convention._c22 is None:
            return

        new_c22 = float(new_c22)

        if self.convention.c22 == new_c22:
            return

        # If factor is changing one has to scale parameters.
        for index in range(len(self._22)):
            self._22[index][3] = self._22[index][3] * self.convention.c22 / new_c22

    def _set_c31(self, new_c31: float) -> None:
        if new_c31 is None or self.convention._c31 is None:
            return

        new_c31 = float(new_c31)

        if self.convention.c31 == new_c31:
            return

        # If factor is changing one has to scale parameters.
        for index in range(len(self._31)):
            self._31[index][1] = self._31[index][1] * self.convention.c31 / new_c31

    def _set_c32(self, new_c32: float) -> None:
        if new_c32 is None or self.convention._c32 is None:
            return

        new_c32 = float(new_c32)

        if self.convention.c32 == new_c32:
            return

        # If factor is changing one has to scale parameters.
        for index in range(len(self._32)):
            self._32[index][3] = self._32[index][3] * self.convention.c32 / new_c32

    def _set_c33(self, new_c33: float) -> None:
        if new_c33 is None or self.convention._c33 is None:
            return

        new_c33 = float(new_c33)

        if self.convention.c33 == new_c33:
            return

        # If factor is changing one has to scale parameters.
        for index in range(len(self._33)):
            self._33[index][5] = self._33[index][5] * self.convention.c33 / new_c33

    def _set_c41(self, new_c41: float) -> None:
        if new_c41 is None or self.convention._c41 is None:
            return

        new_c41 = float(new_c41)

        if self.convention.c41 == new_c41:
            return

        # If factor is changing one has to scale parameters.
        for index in range(len(self._41)):
            self._41[index][1] = self._41[index][1] * self.convention.c41 / new_c41

    def _set_c421(self, new_c421: float) -> None:
        if new_c421 is None or self.convention._c421 is None:
            return

        new_c421 = float(new_c421)

        if self.convention.c421 == new_c421:
            return

        # If factor is changing one has to scale parameters.
        for index in range(len(self._421)):
            self._421[index][3] = self._421[index][3] * self.convention.c421 / new_c421

    def _set_c422(self, new_c422: float) -> None:
        if new_c422 is None or self.convention._c422 is None:
            return

        new_c422 = float(new_c422)

        if self.convention.c422 == new_c422:
            return

        # If factor is changing one has to scale parameters.
        for index in range(len(self._422)):
            self._422[index][3] = self._422[index][3] * self.convention.c422 / new_c422

    def _set_c43(self, new_c43: float) -> None:
        if new_c43 is None or self.convention._c43 is None:
            return

        new_c43 = float(new_c43)

        if self.convention.c43 == new_c43:
            return

        # If factor is changing one has to scale parameters.
        for index in range(len(self._43)):
            self._43[index][5] = self._43[index][5] * self.convention.c43 / new_c43

    def _set_c44(self, new_c44: float) -> None:
        if new_c44 is None or self.convention._c44 is None:
            return

        new_c44 = float(new_c44)

        if self.convention.c44 == new_c44:
            return

        # If factor is changing one has to scale parameters.
        for index in range(len(self._44)):
            self._44[index][7] = self._44[index][7] * self.convention.c44 / new_c44

    ############################################################################
    #                                   Units                                  #
    ############################################################################
    @property
    def units(self) -> str:
        r"""
        Units of the Hamiltonian's parameters.

        .. versionadded:: 0.3.0

        The parameters of the Hamiltonian are stored in some units of energy (or
        energy-like).

        When user adds a parameters to the Hamiltonian (i. e.
        :py:meth:`.SpinHamiltonian.add_21`, ...) the ``parameter`` argument is understood
        to be given in the units of :py:attr:`.SpinHamiltonian.units`.

        By default the Hamiltonian stores and expects parameters in "meV", but the user
        can choose out of the list of the supported units. See
        :ref:`user-guide_usage_units_parameter-units` for the full list of supported units.

        When Hamiltonian already has some parameters in it, then the change of
        :py:attr:`.SpinHamiltonian.units` will convert all parameter to the new units.
        The parameters that the user tries to add afterwards are expected to be in the new
        units already.

        Returns
        -------

        units : str

        See Also
        --------

        :ref:`user-guide_usage_units`
        """

        return _PARAMETER_UNITS_MAKEUP[self._units]

    @units.setter
    def units(self, new_units: str):
        new_units = _validated_units(units=new_units, supported_units=_PARAMETER_UNITS)

        conversion_factor = _PARAMETER_UNITS[self._units] / _PARAMETER_UNITS[new_units]

        # One-site parameters
        for index in range(len(self._1)):
            self._1[index][1] = self._1[index][1] * conversion_factor

        for index in range(len(self._21)):
            self._21[index][1] = self._21[index][1] * conversion_factor

        for index in range(len(self._31)):
            self._31[index][1] = self._31[index][1] * conversion_factor

        for index in range(len(self._41)):
            self._41[index][1] = self._41[index][1] * conversion_factor

        # Two-sites parameters
        for index in range(len(self._22)):
            self._22[index][3] = self._22[index][3] * conversion_factor

        for index in range(len(self._32)):
            self._32[index][3] = self._32[index][3] * conversion_factor

        for index in range(len(self._421)):
            self._421[index][3] = self._421[index][3] * conversion_factor

        for index in range(len(self._422)):
            self._422[index][3] = self._422[index][3] * conversion_factor

        # Three-sites parameters
        for index in range(len(self._33)):
            self._33[index][5] = self._33[index][5] * conversion_factor

        for index in range(len(self._43)):
            self._43[index][5] = self._43[index][5] * conversion_factor

        # Four-sites parameters
        for index in range(len(self._44)):
            self._44[index][7] = self._44[index][7] * conversion_factor

        self._units = new_units.lower()

    ############################################################################
    #                          External magnetic field                         #
    ############################################################################
    # ARGUMENT "h" DEPRECATED since 0.4.0
    # Remove in May of 2026
    def add_magnetic_field(self, B=None, alphas=None, h=None) -> None:
        r"""
        Adds external magnetic field to the Hamiltonian in the form of one spin
        parameters.

        .. math::

            \mu_B  g_{\alpha} \boldsymbol{B}\cdot\boldsymbol{S}_{\mu,\alpha}
            =
            C_1
            \boldsymbol{S}_{\mu,\alpha}
            \cdot
            \boldsymbol{J}_{Zeeman}(\boldsymbol{r}_{\alpha})

        where :math:`\boldsymbol{J}_{Zeeman}(\boldsymbol{r}_{\alpha})` is defined as

        .. math::

            \boldsymbol{J}_{Zeeman}(\boldsymbol{r}_{\alpha})
            =
            \dfrac{\mu_B g_{\alpha}}{C_1}\boldsymbol{B}

        Parameters
        ----------

        B : (3, ) |array-like|_
            Vector of magnetic field (magnetic flux density, B) given in the units of
            Tesla.

        alphas : list of int, optional
            Indices of atoms, to which the magnetic field effect should be added.

        h : (3, ) |array-like|_
            Vector of magnetic field given in the units of Tesla.

            .. deprecated:: 0.4.0
                The argument will be removed in May of 2026. Use ``B`` instead.

        Notes
        -----

        To minimize the energy the magnetic moment will be aligned with the
        direction of the external field. But spin vector will be directed opposite to the
        direction of the magnetic field.

        * If ``alphas is None``, then parameters of the magnetic field added
          only to the magnetic atoms. In other words only to atoms that already have
          at least one other parameter (any) associated with it.
        * If ``alpha is not None``, then parameters of magnetic field are added
          to the atoms with the provided indices (based on the order in
          :py:attr:`.SpinHamiltonian.atoms`)
        """

        if h is not None:
            import warnings

            warnings.warn(
                'Argument "h" is deprecated as of 0.4.0, use "B" instead. "h" will be removed in May of 2026.',
                DeprecationWarning,
                stacklevel=2,
            )
            B = h

        if B is None:
            raise TypeError(
                "SpinHamiltonian.add_magnetic_field() missing 1 required argument: 'B'"
            )

        if self.convention._c1 is None:
            self.convention._c1 = 1.0

        B = np.array(B, dtype=float)

        mu_B = BOHR_MAGNETON / _PARAMETER_UNITS[self._units]  # spinham.units / Tesla

        if alphas is None:
            alphas = self.map_to_all

        zeeman_parameters = [
            mu_B * self.atoms.g_factors[alpha] * B / self.convention.c1
            for alpha in alphas
        ]

        i = 0
        j = 0
        new_p1 = []
        while i < len(alphas) or j < len(self._1):
            if i >= len(alphas) or (j < len(self._1) and alphas[i] > self._1[j][0]):
                new_p1.append(self._1[j])
                j += 1
            elif j >= len(self._1) or (i < len(alphas) and alphas[i] < self._1[j][0]):
                new_p1.append([alphas[i], zeeman_parameters[i]])
                i += 1
            elif alphas[i] == self._1[j][0]:
                new_p1.append(
                    [
                        alphas[i],
                        zeeman_parameters[i] + self._1[j][1],
                    ]
                )
                i += 1
                j += 1

        self._1 = new_p1
        self._reset_internals()

    ############################################################################
    #                    Magnetic dipole-dipole interaction                    #
    ############################################################################

    def add_dipole_dipole(self, R_cut=None, E_cut=None, alphas=None):
        r"""
        Adds magnetic dipole dipole interaction to the Hamiltonian.

        Magnetic dipole dipole interaction is added in the form of two-spin & two-sites
        parameter

        .. math::

            C_{2,2}
            \sum_{\mu,\nu,\alpha,\beta,i,j}
            J_{dd}(\boldsymbol{r}_{\nu,\alpha\beta})^{ij}
            S_{\mu,\alpha}^i
            S_{\mu+\nu,\beta}^j

        where the parameter is defined as

        .. math::

            J_{dd}(\boldsymbol{r}_{\nu,\alpha\beta})^{ij}
            =
            \dfrac{\mu_0\mu_B^2}{8\pi C_{2,2}}
            \dfrac{g_{\alpha}g_{\beta}}{\vert\boldsymbol{r}_{\nu,\alpha\beta}\vert^3}
            (\delta_{k,l} - 3\hat{r}_{\nu,\alpha\beta}^i\hat{r}_{\nu,\alpha\beta}^j)

        if :py:attr:`.SpinHamiltonian.convention.multiple_counting` is ``True`` and as

        .. math::

            J_{dd}(\boldsymbol{r}_{\nu,\alpha\beta})^{ij}
            =
            \dfrac{\mu_0\mu_B^2}{4\pi C_{2,2}}
            \dfrac{g_{\alpha}g_{\beta}}{\vert\boldsymbol{r}_{\nu,\alpha\beta}\vert^3}
            (\delta_{k,l} - 3\hat{r}_{\nu,\alpha\beta}^i\hat{r}_{\nu,\alpha\beta}^j)

        if :py:attr:`.SpinHamiltonian.convention.multiple_counting` is ``False``.

        where :math:`g_{\alpha}` is a g-factor, :math:`\boldsymbol{\hat{r}}_{\nu,\alpha\beta}`
        is a unit vector.

        Parameters
        ----------

        R_cut : float, optional
            Cut off radius for the distance between a pair of atoms.
            :math:`R_{cut} \ge 0`.

        E_cut : float, optional
            Cut off value for the maximum value of the parameter.
            :math:`E_{cut} > 0`.

        alphas : list of int, optional
            Indices of atoms, to which the magnetic field effect should be added.

        Raises
        ------

        ValueError
            * If none of the  ``R_cut`` or ``E_cut`` are provided.
            * If ``R_cut < 0``
            * If ``E_cut <= 0``

        Notes
        -----

        *   If only ``R_cut`` is given, then the dipole dipole term between the pair of
            spins :math:`S_{\mu,\alpha}^i` and :math:`S_{\mu+\nu,\beta}^j` is added if
            :math:`\vert\boldsymbol{r}_{\nu,\alpha\beta}\vert <= R_{cut}`.

        *   If only ``E_cut`` is given, then the ``R_cut`` is estimated as

            .. math::

                R_{cut}
                =
                \left(
                3\sqrt{2}
                \dfrac{\mu_0\mu_B^2g_{max}^2}{8\pi C_{2,2}E_{cut}}
                \right)^{\dfrac{1}{3}}

            if :py:attr:`.SpinHamiltonian.convention.multiple_counting` is ``True`` and as

            .. math::

                R_{cut}
                =
                \left(
                3\sqrt{2}
                \dfrac{\mu_0\mu_B^2g_{max}^2}{4\pi C_{2,2}E_{cut}}
                \right)^{\dfrac{1}{3}}

            if :py:attr:`.SpinHamiltonian.convention.multiple_counting` is ``False``.

            The dipole dipole term between the pair of spins :math:`S_{\mu,\alpha}^i` and
            :math:`S_{\mu+\nu,\beta}^j` is added if
            :math:`\vert\boldsymbol{r}_{\nu,\alpha\beta}\vert \le R_{cut}` and
            :math:`\vert J_{dd}(\boldsymbol{r}_{\nu,\alpha\beta})^{ij}\vert\ge E_{cut}`
            for some :math:`i, j`.

        *   If both ``R_cut`` and ``E_cut`` are provided, then the dipole dipole term
            between the pair of spins :math:`S_{\mu,\alpha}^i` and
            :math:`S_{\mu+\nu,\beta}^j` is added if
            :math:`\vert\boldsymbol{r}_{\nu,\alpha\beta}\vert \le R_{cut}` and
            :math:`\vert J_{dd}(\boldsymbol{r}_{\nu,\alpha\beta})^{ij}\vert \ge E_{cut}`
            for some :math:`i, j`.

        Magnetic dipole-dipole interaction is added either to magnetic atoms or
        to the list of the atoms provided by user.

        * If ``alphas is None``, then parameters of the magnetic field added
          only to the magnetic atoms. In other words only to atoms that already have
          at least one other parameter (any) associated with it.
        * If ``alpha is not None``, then parameters of magnetic field are added
          to the atoms with the provided indices (based on the order in
          :py:attr:`.SpinHamiltonian.atoms`)
        """
        # Constants
        MU_0_MU_B = (
            VACUUM_MAGNETIC_PERMEABILITY
            * BOHR_MAGNETON**2
            / ANGSTROM**3
            / _PARAMETER_UNITS[self._units]
        )  # spinham.units * Angstrom^3

        if E_cut is None and R_cut is None:
            raise ValueError("Expected either E_cut or R_cut, got neither.")

        if E_cut is not None:
            if E_cut <= 0:
                raise ValueError(f"Expected positive cut-off energy, got {E_cut}.")

            R_cut = (
                3
                * np.sqrt(2)
                * MU_0_MU_B
                * max(self.atoms.g_factors) ** 2
                / 4
                / np.pi
                / self.convention.c22
                / E_cut
            )

            if self.convention.multiple_counting:
                R_cut = R_cut / 2

            R_cut = R_cut ** (1 / 3)
        else:
            R_cut = float(R_cut)

        if R_cut < 0:
            raise ValueError(f"Expected positive cut-off radius, got {R_cut}.")

        # Get indices for unit cells of interest
        a1, a2, a3 = self.cell
        a_3_perp = abs(np.cross(a1, a2) @ a3 / np.linalg.norm(np.cross(a1, a2)))
        m_3_max = ceil(R_cut / a_3_perp)
        a_2_perp = np.cross(a2, a3) @ a1 / np.linalg.norm(np.cross(a2, a3))
        m_2_max = ceil(R_cut / a_2_perp)

        m_1_max = ceil(R_cut / np.linalg.norm(a1))

        # Run over all pairs of atoms between (0, 0, 0) and all unit cells of
        # interest
        tmp_parameters = []

        if alphas is None:
            alphas = self.map_to_all

        for alpha in alphas:
            for k in range(-m_3_max, m_3_max + 1):
                for j in range(-m_2_max, m_2_max + 1):
                    for i in range(-m_1_max, m_1_max + 1):
                        for beta in alphas:
                            if k == 0 and j == 0 and i == 0 and alpha == beta:
                                continue

                            if _get_primary_p22(
                                alpha=alpha, beta=beta, nu=(i, j, k)
                            ) != (alpha, beta, (i, j, k)):
                                continue

                            vector = (
                                np.array([i, j, k])
                                + self.atoms.positions[beta]
                                - self.atoms.positions[alpha]
                            ) @ self.cell
                            distance = np.linalg.norm(vector)

                            if distance <= R_cut:
                                parameter = (
                                    MU_0_MU_B
                                    / 4
                                    / np.pi
                                    / self.convention.c22
                                    * self.atoms.g_factors[alpha]
                                    * self.atoms.g_factors[beta]
                                    / distance**3
                                ) * (
                                    np.eye(3, dtype=float)
                                    - 3 * np.outer(vector, vector) / distance**2
                                )
                                if self.convention.multiple_counting:
                                    parameter = parameter / 2

                                if E_cut is None or (np.abs(parameter) >= E_cut).any():
                                    tmp_parameters.append(
                                        [alpha, beta, (i, j, k), parameter]
                                    )

        if len(tmp_parameters) > 0:
            tmp_parameters.sort(key=lambda x: x[:-1])

            self._22 = _merge(list1=self._22, list2=tmp_parameters)

    ############################################################################
    #                                Copy getter                               #
    ############################################################################
    def copy(self):
        R"""
        Returns a new, independent copy of the same Hamiltonian.

        Returns
        -------

        spinham : :py:class:`.SpinHamiltonian`
            A new instance of the same Hamiltonian.
        """

        return deepcopy(self)

    def get_empty(self):
        r"""
        Returns the Hamiltonian with the same cell, atoms, units and convention, but with no
        parameters present.

        Returns
        -------

        spinham : py:class:`.SpinHamiltonian`
            New instance of the spin Hamiltonian.

        Notes
        -----
        Note that in the new Hamiltonian ``spinham.M == 0`` - as there is no parameters
        present, then no atoms are considered to be magnetic.
        """

        return SpinHamiltonian(
            cell=self.cell,
            atoms=self.atoms,
            convention=self.convention,
            units=self.units,
        )

    ############################################################################
    #                           Arithmetic operations                          #
    ############################################################################
    def __mul__(self, number):
        if not isinstance(number, int) and not isinstance(number, float):
            raise TypeError(
                f"unsupported operand type(s) for *: '{type(number)}' and 'SpinHamiltonian'"
            )

        spinham = self.copy()

        # One spin
        for i in range(len(spinham._1)):
            spinham._1[i][1] *= number

        # Two spins
        for i in range(len(spinham._21)):
            spinham._21[i][1] *= number

        for i in range(len(spinham._22)):
            spinham._22[i][3] *= number

        # Three spins
        for i in range(len(spinham._31)):
            spinham._31[i][1] *= number

        for i in range(len(spinham._32)):
            spinham._32[i][3] *= number

        for i in range(len(spinham._33)):
            spinham._33[i][5] *= number

        # Four spins
        for i in range(len(spinham._41)):
            spinham._41[i][1] *= number

        for i in range(len(spinham._421)):
            spinham._421[i][3] *= number

        for i in range(len(spinham._422)):
            spinham._422[i][3] *= number

        for i in range(len(spinham._43)):
            spinham._43[i][5] *= number

        for i in range(len(spinham._44)):
            spinham._44[i][7] *= number

        return spinham

    def __rmul__(self, number):
        return self.__mul__(number=number)

    def __add__(self, other):
        if not isinstance(other, SpinHamiltonian):
            raise NotImplementedError

        # Check that unit cells are the same
        if not np.allclose(self.cell, other.cell):
            raise ValueError(
                "Unit cells of two Hamiltonians are different, "
                "summation is not supported"
            )

        # Check that atoms are the same
        same_atoms = True
        if len(self.atoms.names) != len(other.atoms.names):
            same_atoms = False
        else:
            for i in range(len(self.atoms.names)):
                if (
                    self.atoms.names[i] != other.atoms.names[i]
                    or not np.allclose(
                        self.atoms.positions[i], other.atoms.positions[i]
                    )
                    or abs(self.atoms.spins[i] - other.atoms.spins[i]) > 1e-8
                    or abs(self.atoms.g_factors[i] - other.atoms.g_factors[i]) > 1e-8
                ):
                    same_atoms = False

        if not same_atoms:
            raise ValueError(
                "Atoms of two spin Hamiltonians are different, "
                "summation is not supported."
            )

        # Make sure that units are the same
        other_units = other.units
        other.units = self.units

        # Make sure that conventions are the same
        other_convention = other.convention
        other.convention = self.convention

        result = self.get_empty()

        # One spin terms
        result._1 = _merge(list1=self._1, list2=other._1)

        # Two spin terms
        result._21 = _merge(list1=self._21, list2=other._21)
        result._22 = _merge(list1=self._22, list2=other._22)

        # Three spin terms
        result._31 = _merge(list1=self._31, list2=other._31)
        result._32 = _merge(list1=self._32, list2=other._32)
        result._33 = _merge(list1=self._33, list2=other._33)

        # Four spin terms
        result._41 = _merge(list1=self._41, list2=other._41)
        result._421 = _merge(list1=self._421, list2=other._421)
        result._422 = _merge(list1=self._422, list2=other._422)
        result._43 = _merge(list1=self._43, list2=other._43)
        result._44 = _merge(list1=self._44, list2=other._44)

        # Restore units of other Hamiltonian
        other.units = other_units

        # Restore convention of other Hamiltonian
        other.convention = other_convention

        return result

    def __sub__(self, other):
        return self + (-1) * other

    ############################################################################
    #                            One spin & one site                           #
    ############################################################################
    p1 = _p1
    add_1 = _add_1
    remove_1 = _remove_1

    ############################################################################
    #                           Two spins & one site                           #
    ############################################################################
    p21 = _p21
    add_21 = _add_21
    remove_21 = _remove_21

    ############################################################################
    #                           Two spins & two sites                          #
    ############################################################################
    p22 = _p22
    add_22 = _add_22
    remove_22 = _remove_22

    ############################################################################
    #                          Three spins & one site                          #
    ############################################################################
    p31 = _p31
    add_31 = _add_31
    remove_31 = _remove_31

    ############################################################################
    #                          Three spins & two sites                         #
    ############################################################################
    p32 = _p32
    add_32 = _add_32
    remove_32 = _remove_32

    ############################################################################
    #                         Three spins & three sites                        #
    ############################################################################
    p33 = _p33
    add_33 = _add_33
    remove_33 = _remove_33

    ############################################################################
    #                           Four spins & one site                          #
    ############################################################################
    p41 = _p41
    add_41 = _add_41
    remove_41 = _remove_41

    ############################################################################
    #                          Four spins & two sites (3+1)                    #
    ############################################################################
    p421 = _p421
    add_421 = _add_421
    remove_421 = _remove_421

    ############################################################################
    #                          Four spins & two sites (2+2)                    #
    ############################################################################
    p422 = _p422
    add_422 = _add_422
    remove_422 = _remove_422

    ############################################################################
    #                         Four spins & three sites                         #
    ############################################################################
    p43 = _p43
    add_43 = _add_43
    remove_43 = _remove_43

    ############################################################################
    #                          Four spins & four sites                         #
    ############################################################################
    p44 = _p44
    add_44 = _add_44
    remove_44 = _remove_44


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================


import numpy as np
from numpy.linalg import LinAlgError

from magnopy._exceptions import ColpaFailed

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def _check_grand_dynamical_matrix(D):
    r"""
    Check that grand dynamical matrix is (2N, 2N) matrix

    Parameters
    ----------

    D : |array-like|_
        Candidate for the grand dynamical matrix

    Returns
    -------

    D : (2N, 2N) :numpy:`ndarray`
        Grand dynamical matrix.

    N : int

    Raises
    ------

    ValueError
        If the check is not passed
    """

    D = np.array(D)

    if len(D.shape) != 2:
        raise ValueError(f"Grand dynamical matrix is not 2-dimensional, got {D.shape}.")

    if D.shape[0] != D.shape[1]:
        raise ValueError(f"Grand dynamical matrix is not square, got {D.shape}.")

    if D.shape[0] % 2 != 0:
        raise ValueError(
            f"Size of the grand dynamical matrix is not even, got {D.shape}."
        )

    return D, D.shape[0] // 2


def _inverse_by_colpa(matrix):
    # Compute G from G^-1 (or vise versa) following Colpa, see equation (3.7) for details

    N = matrix.shape[0] // 2
    matrix = np.conjugate(matrix).T
    matrix[:N, N:] *= -1
    matrix[N:, :N] *= -1

    return matrix


def solve_via_colpa(D):
    r"""
    Diagonalizes grand-dynamical matrix following the method of Colpa.

    An algorithm is described in section 3, remark 1 of [1]_.

    Solves the Bogoliubov Hamiltonian of the form

    .. math::

        \hat{H} = \sum_{r^{\prime}, r = 1}^m
        a_{r^{\prime}}^{\dagger}\boldsymbol{\Delta}_1^{r^{\prime}r}a_r +
        a_{r^{\prime}}^{\dagger}\boldsymbol{\Delta}_2^{r^{\prime}r}a_{m+r}^{\dagger} +
        a_{m+r^{\prime}}^{\dagger}\boldsymbol{\Delta}_3^{r^{\prime}r}a_r +
        a_{m+r^{\prime}}^{\dagger}\boldsymbol{\Delta}_4^{r^{\prime}r}a_{m+r}^{\dagger}

    ensuring the bosonic commutation relations.

    In a matrix form the Hamiltonian can be written as

    .. math::

        \hat{H} = \boldsymbol{\cal A}^{\dagger} \boldsymbol{D} \boldsymbol{\cal A}

    where

    .. math::

        \boldsymbol{\cal A} =
        \begin{pmatrix}
            a_1 \\
            \cdots \\
            a_m \\
            a_{m+1} \\
            \cdots \\
            a_{2m} \\
        \end{pmatrix}

    After diagonalization the Hamiltonian has the form

    .. math::

        \hat{H}
        =
        \boldsymbol{\cal B}^{\dagger} \boldsymbol{\mathcal{E}} \boldsymbol{\cal B}

    where

    .. math::

        \boldsymbol{\cal B} =
        \begin{pmatrix}
            b_1 \\
            \cdots \\
            b_m \\
            b_{m+1} \\
            \cdots \\
            b_{2m} \\
        \end{pmatrix}

    Parameters
    ----------

    D : (2N, 2N) |array-like|_
        Grand dynamical matrix. If it is Hermitian and positive-defined, then obtained
        eigenvalues are positive and real.

        .. math::

            \boldsymbol{\mathcal{D}} = \begin{pmatrix}
                \boldsymbol{\Delta_1} & \boldsymbol{\Delta_2} \\
                \boldsymbol{\Delta_3} & \boldsymbol{\Delta_4}
            \end{pmatrix}

    Returns
    -------
    E : (2N,) :numpy:`ndarray`
        The eigenvalues. It is an array of the diagonal elements of the diagonal matrix
        :math:`\boldsymbol{\mathcal{E}}`. First N elements correspond to the
        :math:`b^{\dagger}_1b_1, \dots, b^{\dagger}_mb_m` and last N elements - to
        the :math:`b^{\dagger}_{m+1}b_{m+1}, \dots, b^{\dagger}_{2m}b_{2m}`.

        First N eigenvalues are sorted in descending order, while last N eigenvalues are
        sorted in ascending order.

        .. math::

            \boldsymbol{\mathcal{E}}
            =
            (\boldsymbol{G}^{\dagger})^{-1} \boldsymbol{D} \boldsymbol{G}^{-1}

    G : (2N, 2N) :numpy:`ndarray`
        Transformation matrix, that changes the basis from the original set of bosonic
        operators :math:`\boldsymbol{a}` to the set of new bosonic operators
        :math:`\boldsymbol{b}` which diagonalize the Hamiltonian:

        .. math::
            \boldsymbol{\cal B} = \boldsymbol{G} \boldsymbol{\cal A}

        The rows are ordered in the same way as the eigenvalues.

    Raises
    ------

    ColpaFailed
        If the algorithm fails. Typically it means that the grand dynamical matrix
        :math:`\boldsymbol{D}` is not positive-defined.

    ValueError
        If the grand dynamical matrix is not square or its shape is not even.

    References
    ----------

    .. [1] Colpa, J.H.P., 1978.
        Diagonalization of the quadratic boson hamiltonian.
        Physica A: Statistical Mechanics and its Applications,
        93(3-4), pp.327-353.

    Examples
    --------

    For already diagonal matrix this function does not do much

    .. doctest::

        >>> import magnopy
        >>> D = [[1, 0], [0, 2]]
        >>> E, G = magnopy.solve_via_colpa(D)
        >>> E
        array([1., 2.])
        >>> G
        array([[ 1., -0.],
               [-0.,  1.]])

    .. doctest::

        >>> import magnopy
        >>> D = [[1, 1j], [-1j, 2]]
        >>> E, G = magnopy.solve_via_colpa(D)
        >>> E
        array([0.61803399+0.j, 1.61803399+0.j])
        >>> G
        array([[ 1.08204454-0.j        ,  0.        +0.41330424j],
               [-0.        -0.41330424j,  1.08204454-0.j        ]])
        >>> E, G = magnopy.solve_via_colpa(D)
        >>> E
        array([0.61803399+0.j, 1.61803399+0.j])
        >>> G # doctest: +SKIP
        array([[1.08204454+0.j        , 0.        -0.41330424j],
               [0.        +0.41330424j, 1.08204454+0.j        ]])
    """

    # Guarantee that D is 2Nx2N matrix
    D, N = _check_grand_dynamical_matrix(D)

    # Para-unitary matrix
    g = np.diag(np.concatenate((np.ones(N), -np.ones(N))))

    # Cholesky decomposition D = K^{\dag}K
    try:
        # In Colpa article decomposition is K^{\dag}K,
        # while numpy gives KK^{\dag} by default,
        # thus upper=True is needed
        K = np.linalg.cholesky(D, upper=True)
    except LinAlgError:
        raise ColpaFailed

    # Solve the standard eigenvalue problem for K g K^{\dag}
    L, U = np.linalg.eig(K @ g @ np.conjugate(K).T)

    # Sort with respect to L, in descending order

    # Step 1 put L as a first line on top of U:
    # [
    #   [L_0, ..., L_2N],
    #   [U_0_0, ..., U_0_2N],
    #   ...
    #   [U_2N_0, ..., U_2N_2N],
    # ]
    U = np.concatenate((L[np.newaxis, :], U), axis=0)

    # Step 2 sort each row according to the sorted order of the first row (L)
    U = U[:, np.argsort(U[0])]

    # Step 3 separate L and U, and reverse the order to have
    # first N eigenvalues to be positive
    # and last N eigenvalues to be negative
    L = U[0, ::-1]
    U = U[1:, ::-1]

    # Compute actual eigenvalues of D
    E = g @ L

    # Compute G_inv
    G_inv = np.linalg.inv(K) @ U @ np.sqrt(np.diag(E))

    # Inverse it by the Colpa's method to get G
    G = _inverse_by_colpa(G_inv)

    return E, G


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================


import numpy as np

from magnopy._diagonalization import solve_via_colpa
from magnopy._exceptions import ColpaFailed
from magnopy._local_rf import span_local_rfs

from magnopy._data_validation import _validated_units
from magnopy._constants._units import _ENERGY_UNITS, _MAGNON_ENERGY_UNITS


# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


class LSWT:
    r"""
    Linear Spin Wave theory.

    It is created from some spin Hamiltonian and set of direction vectors, that defines
    the ground state.

    Parameters
    ----------

    spinham : :py:class:`.SpinHamiltonian`
        Spin Hamiltonian.

    spin_directions : (M, 3) |array-like|_
        Directions of spin vectors. Only directions of vectors are used, modulus is
        ignored. If spin Hamiltonian contains non-magnetic atom, then only the spin
        directions for the magnetic atoms are expected. The order of spin directions is
        the same as the order of magnetic atoms in
        :py:attr:`SpinHamiltonian.magnetic_atoms`. See Notes for more details.

    Attributes
    ----------

    z : (M, 3) :numpy:`ndarray`
        Spin directions (directions of local quantization axes).

    p : (M, 3) :numpy:`ndarray`
        Hybridized x and y components of the local coordinate system
        :math:`\mathbf{p} = \mathbf{x} + i \mathbf{y}`.

    M : int
        Number of spins in the unit cell

    cell : (3, 3) :numpy:`ndarray`
        Unit cell. Rows are vectors, columns are cartesian components.

    spins : (M, ) :numpy:`ndarray`
        Spin values of the magnetic centers.

    Notes
    -----

    Let the spin Hamiltonian contain three atoms Cr1, Br, Cr3 in that order. Assume that
    two atoms are magnetic (Cr1 and Cr3), one atom is not (Br). Then ``spin_directions``
    is a (2, 3) array with ``spin_directions[0]`` being the direction for spin of Cr1 and
    ``spin_directions[1]`` being the direction of spin for Cr3.

    Examples
    --------

    .. doctest::

        >>> import magnopy
        >>> spinham = magnopy.examples.cubic_ferro_nn()
        >>> lswt = magnopy.LSWT(spinham=spinham, spin_directions=[[0, 0, 1]])
    """

    def __init__(self, spinham, spin_directions):
        spin_directions = np.array(spin_directions, dtype=float)
        spin_directions /= np.linalg.norm(spin_directions, axis=1)[:, np.newaxis]

        x, y, self.z = span_local_rfs(
            directional_vectors=spin_directions, hybridize=False
        )
        self.p = x + 1j * y

        self.spins = np.array(spinham.magnetic_atoms.spins, dtype=float)

        initial_units = spinham.units
        initial_convention = spinham.convention

        spinham.units = "mev"
        spinham.convention = initial_convention.get_modified(
            spin_normalized=False, multiple_counting=True
        )

        self.M = spinham.M
        self.cell = spinham.cell

        ########################################################################
        #                    Renormalized one-spin parameter                   #
        ########################################################################
        self._J1 = np.zeros((self.M, 3), dtype=float)

        # One spin
        for alpha, parameter in spinham.p1:
            alpha = spinham.map_to_magnetic[alpha]
            self._J1[alpha] = self._J1[alpha] + spinham.convention.c1 * parameter

        # Two spins & one site
        for alpha, parameter in spinham.p21:
            alpha = spinham.map_to_magnetic[alpha]
            self._J1[alpha] = self._J1[alpha] + (
                2
                * spinham.convention.c21
                * np.einsum("ij,j->i", parameter, self.z[alpha])
                * self.spins[alpha]
            )

        # Two spins & two sites
        for alpha, beta, _, parameter in spinham.p22:
            alpha = spinham.map_to_magnetic[alpha]
            beta = spinham.map_to_magnetic[beta]
            self._J1[alpha] = self._J1[alpha] + (
                2
                * spinham.convention.c22
                * (parameter @ self.z[beta])
                * self.spins[beta]
            )

        # Three spins & one site
        for alpha, parameter in spinham.p31:
            alpha = spinham.map_to_magnetic[alpha]
            self._J1[alpha] = self._J1[alpha] + (
                3
                * spinham.convention.c31
                * np.einsum("iju,j,u->i", parameter, self.z[alpha], self.z[alpha])
                * self.spins[alpha] ** 2
            )

        # Three spins & two sites
        for alpha, beta, _, parameter in spinham.p32:
            alpha = spinham.map_to_magnetic[alpha]
            beta = spinham.map_to_magnetic[beta]
            self._J1[alpha] = self._J1[alpha] + (
                3
                * spinham.convention.c32
                * np.einsum("iju,j,u->i", parameter, self.z[alpha], self.z[beta])
                * self.spins[alpha]
                * self.spins[beta]
            )

        # Three spins & three sites
        for alpha, beta, gamma, _, _, parameter in spinham.p33:
            alpha = spinham.map_to_magnetic[alpha]
            beta = spinham.map_to_magnetic[beta]
            gamma = spinham.map_to_magnetic[gamma]
            self._J1[alpha] = self._J1[alpha] + (
                3
                * spinham.convention.c33
                * np.einsum("iju,j,u->i", parameter, self.z[beta], self.z[gamma])
                * self.spins[beta]
                * self.spins[gamma]
            )

        # Four spins & one site
        for alpha, parameter in spinham.p41:
            alpha = spinham.map_to_magnetic[alpha]
            self._J1[alpha] = self._J1[alpha] + (
                4
                * spinham.convention.c41
                * np.einsum(
                    "ijuv,j,u,v->i",
                    parameter,
                    self.z[alpha],
                    self.z[alpha],
                    self.z[alpha],
                )
                * self.spins[alpha] ** 3
            )

        # Four spins & two sites (1+3)
        for alpha, beta, _, parameter in spinham.p421:
            alpha = spinham.map_to_magnetic[alpha]
            beta = spinham.map_to_magnetic[beta]
            self._J1[alpha] = self._J1[alpha] + (
                4
                * spinham.convention.c421
                * np.einsum(
                    "ijuv,j,u,v->i",
                    parameter,
                    self.z[alpha],
                    self.z[alpha],
                    self.z[beta],
                )
                * self.spins[alpha] ** 2
                * self.spins[beta]
            )

        # Four spins & two sites (2+2)
        for alpha, beta, _, parameter in spinham.p422:
            alpha = spinham.map_to_magnetic[alpha]
            beta = spinham.map_to_magnetic[beta]
            self._J1[alpha] = self._J1[alpha] + (
                4
                * spinham.convention.c422
                * np.einsum(
                    "ijuv,j,u,v->i",
                    parameter,
                    self.z[alpha],
                    self.z[beta],
                    self.z[beta],
                )
                * self.spins[alpha]
                * self.spins[beta] ** 2
            )

        # Four spins & three sites
        for alpha, beta, gamma, _, _, parameter in spinham.p43:
            alpha = spinham.map_to_magnetic[alpha]
            beta = spinham.map_to_magnetic[beta]
            gamma = spinham.map_to_magnetic[gamma]
            self._J1[alpha] = self._J1[alpha] + (
                4
                * spinham.convention.c43
                * np.einsum(
                    "ijuv,j,u,v->i",
                    parameter,
                    self.z[alpha],
                    self.z[beta],
                    self.z[gamma],
                )
                * self.spins[alpha]
                * self.spins[beta]
                * self.spins[gamma]
            )

        # Four spins & four sites
        for alpha, beta, gamma, epsilon, _, _, _, parameter in spinham.p44:
            alpha = spinham.map_to_magnetic[alpha]
            beta = spinham.map_to_magnetic[beta]
            gamma = spinham.map_to_magnetic[gamma]
            epsilon = spinham.map_to_magnetic[epsilon]
            self._J1[alpha] = self._J1[alpha] + (
                4
                * spinham.convention.c44
                * np.einsum(
                    "ijuv,j,u,v->i",
                    parameter,
                    self.z[beta],
                    self.z[gamma],
                    self.z[epsilon],
                )
                * self.spins[beta]
                * self.spins[gamma]
                * self.spins[epsilon]
            )

        ########################################################################
        #                   Renormalized two-spins parameter                   #
        ########################################################################
        self._J2 = {}

        # First - terms with delta in from of them

        # Two spins & one site
        for alpha, parameter in spinham.p21:
            alpha = spinham.map_to_magnetic[alpha]
            if (0, 0, 0) not in self._J2:
                self._J2[(0, 0, 0)] = np.zeros((self.M, self.M, 3, 3), dtype=float)

            self._J2[(0, 0, 0)][alpha, alpha] += spinham.convention.c21 * parameter

        # Three spins & one site
        for alpha, parameter in spinham.p31:
            alpha = spinham.map_to_magnetic[alpha]
            if (0, 0, 0) not in self._J2:
                self._J2[(0, 0, 0)] = np.zeros((self.M, self.M, 3, 3), dtype=float)

            self._J2[(0, 0, 0)][alpha, alpha] += (
                3
                * spinham.convention.c31
                * np.einsum("iju,u->ij", parameter, self.z[alpha])
                * self.spins[alpha]
            )

        # Four spins & one site
        for alpha, parameter in spinham.p41:
            alpha = spinham.map_to_magnetic[alpha]
            if (0, 0, 0) not in self._J2:
                self._J2[(0, 0, 0)] = np.zeros((self.M, self.M, 3, 3), dtype=float)

            self._J2[(0, 0, 0)][alpha, alpha] += (
                6
                * spinham.convention.c41
                * np.einsum("ijuv,u,v->ij", parameter, self.z[alpha], self.z[alpha])
                * self.spins[alpha] ** 2
            )

        # Then all other parameters

        # Two spins & two sites
        for alpha, beta, nu, parameter in spinham.p22:
            alpha = spinham.map_to_magnetic[alpha]
            beta = spinham.map_to_magnetic[beta]
            if nu not in self._J2:
                self._J2[nu] = np.zeros((self.M, self.M, 3, 3), dtype=float)

            self._J2[nu][alpha, beta] += spinham.convention.c22 * parameter

        # Three spins & two sites
        for alpha, beta, nu, parameter in spinham.p32:
            alpha = spinham.map_to_magnetic[alpha]
            beta = spinham.map_to_magnetic[beta]
            if nu not in self._J2:
                self._J2[nu] = np.zeros((self.M, self.M, 3, 3), dtype=float)

            self._J2[nu][alpha, beta] += (
                3
                * spinham.convention.c32
                * np.einsum("iuj,u->ij", parameter, self.z[alpha])
                * self.spins[alpha]
            )

        # Three spins & three sites
        for alpha, beta, gamma, nu, _, parameter in spinham.p33:
            alpha = spinham.map_to_magnetic[alpha]
            beta = spinham.map_to_magnetic[beta]
            gamma = spinham.map_to_magnetic[gamma]
            if nu not in self._J2:
                self._J2[nu] = np.zeros((self.M, self.M, 3, 3), dtype=float)

            self._J2[nu][alpha, beta] += (
                3
                * spinham.convention.c33
                * np.einsum("iju,u->ij", parameter, self.z[gamma])
                * self.spins[gamma]
            )

        # Four spins & two sites (1+3)
        for alpha, beta, nu, parameter in spinham.p421:
            alpha = spinham.map_to_magnetic[alpha]
            beta = spinham.map_to_magnetic[beta]
            if nu not in self._J2:
                self._J2[nu] = np.zeros((self.M, self.M, 3, 3), dtype=float)

            self._J2[nu][alpha, beta] += (
                6
                * spinham.convention.c421
                * np.einsum("iuvj,u,v->ij", parameter, self.z[alpha], self.z[alpha])
                * self.spins[alpha] ** 2
            )

        # Four spins & two sites (2+2)
        for alpha, beta, nu, parameter in spinham.p422:
            alpha = spinham.map_to_magnetic[alpha]
            beta = spinham.map_to_magnetic[beta]
            if nu not in self._J2:
                self._J2[nu] = np.zeros((self.M, self.M, 3, 3), dtype=float)

            self._J2[nu][alpha, beta] += (
                6
                * spinham.convention.c422
                * np.einsum("iujv,u,v->ij", parameter, self.z[alpha], self.z[beta])
                * self.spins[alpha]
                * self.spins[beta]
            )

        # Four spins & three sites
        for alpha, beta, gamma, nu, _, parameter in spinham.p43:
            alpha = spinham.map_to_magnetic[alpha]
            beta = spinham.map_to_magnetic[beta]
            gamma = spinham.map_to_magnetic[gamma]
            if nu not in self._J2:
                self._J2[nu] = np.zeros((self.M, self.M, 3, 3), dtype=float)

            self._J2[nu][alpha, beta] += (
                6
                * spinham.convention.c43
                * np.einsum("iujv,u->ij", parameter, self.z[alpha], self.z[gamma])
                * self.spins[alpha]
                * self.spins[gamma]
            )

        # Four spins & four sites
        for alpha, beta, gamma, epsilon, nu, _, _, parameter in spinham.p44:
            alpha = spinham.map_to_magnetic[alpha]
            beta = spinham.map_to_magnetic[beta]
            gamma = spinham.map_to_magnetic[gamma]
            epsilon = spinham.map_to_magnetic[epsilon]
            if nu not in self._J2:
                self._J2[nu] = np.zeros((self.M, self.M, 3, 3), dtype=float)

            self._J2[nu][alpha, beta] += (
                6
                * spinham.convention.c44
                * np.einsum("ijuv,u->ij", parameter, self.z[gamma], self.z[epsilon])
                * self.spins[gamma]
                * self.spins[epsilon]
            )

        spinham.units = initial_units
        spinham.convention = initial_convention

        self.A1 = 0.5 * np.sum(self._J1 * self.z, axis=1)

        self.A2 = {}
        self.B2 = {}

        for nu in self._J2:
            self.A2[nu] = 0.5 * np.einsum(
                "abij,a,b,ai,bj->ab",
                self._J2[nu],
                np.sqrt(self.spins),
                np.sqrt(self.spins),
                self.p,
                np.conjugate(self.p),
            )
            self.B2[nu] = 0.5 * np.einsum(
                "abij,a,b,ai,bj->ab",
                self._J2[nu],
                np.sqrt(self.spins),
                np.sqrt(self.spins),
                np.conjugate(self.p),
                np.conjugate(self.p),
            )

    def E_2(self, units="meV") -> float:
        r"""
        Computes the correction to the ground state energy that arises from the LSWT.

        Parameters
        ----------

        units : str, default "meV"
            .. versionadded:: 0.3.0

            Units of energy. See :ref:`user-guide_usage_units_energy-units` for the full
            list of supported units.

        Returns
        -------

        E_2 : float

        Examples
        --------

        .. doctest::

            >>> import magnopy
            >>> spinham = magnopy.examples.cubic_ferro_nn()
            >>> lswt = magnopy.LSWT(spinham=spinham, spin_directions=[[0, 0, 1]])
            >>> # Default units are meV
            >>> lswt.E_2()
            -1.5
        """

        result = float(0.5 * np.sum(self._J1 * self.z))

        # Convert units if necessary
        if units != "meV":
            units = _validated_units(units=units, supported_units=_ENERGY_UNITS)
            result = result * _ENERGY_UNITS["mev"] / _ENERGY_UNITS[units]

        return result

    def O(self, units="meV"):  # noqa E743
        r"""
        Computes coefficient of the one-operator terms.

        Parameters
        ----------

        units : str, default "meV"
            .. versionadded:: 0.3.0

            Units of energy. See :ref:`user-guide_usage_units_energy-units` for the full
            list of supported units.


        Returns
        -------

        O : (M, ) :numpy:`ndarray`
            Elements are complex numbers.

        Notes
        -----

        Before the diagonalization, the magnon Hamiltonian has the form

        .. math::

            \mathcal{H}
            =
            \dots
            +
            \sqrt{N}
            \sum_{\alpha}
            \Bigl(
            O_{\alpha}
            a_{\alpha}(\boldsymbol{0})
            +
            \overline{O_{\alpha}}
            a^{\dagger}_{\alpha}(\boldsymbol{0})
            \Bigr)
            +
            \dots

        where overline denotes complex conjugation. This function computes the
        coefficients :math:`O_{\alpha}`.

        Examples
        --------

        .. doctest::

            >>> import magnopy
            >>> spinham = magnopy.examples.cubic_ferro_nn()
            >>> lswt = magnopy.LSWT(spinham=spinham, spin_directions=[[0, 0, 1]])
            >>> lswt.O()
            array([0.+0.j])
        """

        result = np.einsum(
            "a,ai,ai->a",
            np.sqrt(self.spins) / np.sqrt(2),
            np.conjugate(self.p),
            self._J1,
        )

        # Convert units if necessary
        if units != "meV":
            units = _validated_units(units=units, supported_units=_ENERGY_UNITS)
            result = result * _ENERGY_UNITS["mev"] / _ENERGY_UNITS[units]

        return result

    def A(self, k, relative=False, units="meV"):
        r"""
        Computes part of the grand dynamical matrix.

        Parameters
        ----------

        k : (3,) |array-like|_
            Reciprocal vector

        relative : bool, default False
            If ``relative=True``, then ``k`` is interpreted as given relative to the
            reciprocal unit cell. Otherwise it is interpreted as given in absolute
            coordinates.

        units : str, default "meV"
            .. versionadded:: 0.3.0

            Units of energy. See :ref:`user-guide_usage_units_energy-units` for the full
            list of supported units.


        Returns
        -------

        A : (M, M) :numpy:`ndarray`
            :math:`A_{\alpha\beta}(\boldsymbol{k})`.

        Notes
        -----

        Before the diagonalization, the magnon Hamiltonian has the form

        .. math::

            \mathcal{H}
            =
            \dots
            +
            \sum_{\boldsymbol{k}, \alpha}
            \boldsymbol{\mathcal{A}}(\boldsymbol{k})^{\dagger}
            \begin{pmatrix}
            \boldsymbol{A}(\boldsymbol{k}) & \boldsymbol{B}^{\dagger}(\boldsymbol{k}) \\
            \boldsymbol{B}(\boldsymbol{k}) & \overline{\boldsymbol{A}(-\boldsymbol{k})}
            \end{pmatrix}
            \boldsymbol{\mathcal{A}}(\boldsymbol{k})

        where

        .. math::
            \boldsymbol{\mathcal{A}}(\boldsymbol{k})
            =
            \begin{pmatrix}
            a_1(\boldsymbol{k}),
            \dots,
            a_M(\boldsymbol{k}),
            a^{\dagger}_1(-\boldsymbol{k}),
            \dots,
            a^{\dagger}_M(-\boldsymbol{k}),
            \end{pmatrix}

        This function computes the matrix :math:`\boldsymbol{A}(\boldsymbol{k})`.

        See Also
        --------

        LSWT.B
        LSWT.GDM

        Examples
        --------

        .. doctest::

            >>> import magnopy
            >>> spinham = magnopy.examples.cubic_ferro_nn()
            >>> lswt = magnopy.LSWT(spinham=spinham, spin_directions=[[0, 0, 1]])
            >>> lswt.A(k=[0, 0, 0.5], relative=True)
            array([[1.+0.j]])
        """

        k = np.array(k)

        result = np.zeros((self.M, self.M), dtype=complex)

        for nu in self.A2:
            if relative:
                phase = 2 * np.pi * (k @ nu)
            else:
                phase = k @ (nu @ self.cell)
            result = result + self.A2[nu] * np.exp(1j * phase)

        result = result - np.diag(self.A1)

        # Convert units if necessary
        if units != "meV":
            units = _validated_units(units=units, supported_units=_ENERGY_UNITS)
            result = result * _ENERGY_UNITS["mev"] / _ENERGY_UNITS[units]

        return result

    def B(self, k, relative=False, units="meV"):
        r"""
        Computes part of the grand dynamical matrix.

        Parameters
        ----------

        k : (3,) |array-like|_
            Reciprocal vector.

        relative : bool, default False
            If ``relative=True``, then ``k`` is interpreted as given relative to the
            reciprocal unit cell. Otherwise it is interpreted as given in absolute
            coordinates.

        units : str, default "meV"
            .. versionadded:: 0.3.0

            Units of energy. See :ref:`user-guide_usage_units_energy-units` for the full
            list of supported units.


        Returns
        -------

        B : (M, M) :numpy:`ndarray`
            :math:`B_{\alpha\beta}(\boldsymbol{k})`.

        Notes
        -----

        Before the diagonalization, the magnon Hamiltonian has the form

        .. math::

            \mathcal{H}
            =
            \dots
            +
            \sum_{\boldsymbol{k}, \alpha}
            \boldsymbol{\mathcal{A}}(\boldsymbol{k})^{\dagger}
            \begin{pmatrix}
            \boldsymbol{A}(\boldsymbol{k}) & \boldsymbol{B}^{\dagger}(\boldsymbol{k}) \\
            \boldsymbol{B}(\boldsymbol{k}) & \overline{\boldsymbol{A}(-\boldsymbol{k})}
            \end{pmatrix}
            \boldsymbol{\mathcal{A}}(\boldsymbol{k})

        where

        .. math::
            \boldsymbol{\mathcal{A}}(\boldsymbol{k})
            =
            \begin{pmatrix}
            a_1(\boldsymbol{k}),
            \dots,
            a_M(\boldsymbol{k}),
            a^{\dagger}_1(-\boldsymbol{k}),
            \dots,
            a^{\dagger}_M(-\boldsymbol{k}),
            \end{pmatrix}

        This function computes the matrix :math:`\boldsymbol{B}(\boldsymbol{k})`.

        See Also
        --------

        LSWT.A
        LSWT.GDM

        Examples
        --------

        .. doctest::

            >>> import magnopy
            >>> spinham = magnopy.examples.cubic_ferro_nn()
            >>> lswt = magnopy.LSWT(spinham=spinham, spin_directions=[[0, 0, 1]])
            >>> lswt.B(k=[0, 0, 0.5], relative=True)
            array([[0.+0.j]])
        """

        k = np.array(k)

        result = np.zeros((self.M, self.M), dtype=complex)

        for nu in self.B2:
            if relative:
                phase = 2 * np.pi * (k @ nu)
            else:
                phase = k @ (nu @ self.cell)
            result = result + self.B2[nu] * np.exp(1j * phase)

        # Convert units if necessary
        if units != "meV":
            units = _validated_units(units=units, supported_units=_ENERGY_UNITS)
            result = result * _ENERGY_UNITS["mev"] / _ENERGY_UNITS[units]

        return result

    def GDM(self, k, relative=False, units="meV"):
        r"""
        Computes grand dynamical matrix.

        Parameters
        ----------

        k : (3,) |array-like|_
            Reciprocal vector.

        relative : bool, default False
            If ``relative=True``, then ``k`` is interpreted as given relative to the
            reciprocal unit cell. Otherwise it is interpreted as given in absolute
            coordinates.

        units : str, default "meV"
            .. versionadded:: 0.3.0

            Units of energy. See :ref:`user-guide_usage_units_energy-units` for the full
            list of supported units.


        Returns
        -------

        gdm : (2M, 2M) :numpy:`ndarray`
            Gran dynamical matrix.

        Notes
        -----

        Before the diagonalization, the magnon Hamiltonian has the form

        .. math::

            \mathcal{H}
            =
            \dots
            +
            \sum_{\boldsymbol{k}, \alpha}
            \boldsymbol{\mathcal{A}}(\boldsymbol{k})^{\dagger}
            \begin{pmatrix}
            \boldsymbol{A}(\boldsymbol{k}) & \boldsymbol{B}^{\dagger}(\boldsymbol{k}) \\
            \boldsymbol{B}(\boldsymbol{k}) & \overline{\boldsymbol{A}(-\boldsymbol{k})}
            \end{pmatrix}
            \boldsymbol{\mathcal{A}}(\boldsymbol{k})

        where

        .. math::
            \boldsymbol{\mathcal{A}}(\boldsymbol{k})
            =
            \begin{pmatrix}
            a_1(\boldsymbol{k}),
            \dots,
            a_M(\boldsymbol{k}),
            a^{\dagger}_1(-\boldsymbol{k}),
            \dots,
            a^{\dagger}_M(-\boldsymbol{k}),
            \end{pmatrix}

        This function computes the grand dynamical matrix
        :math:`\boldsymbol{D}(\boldsymbol{k})`

        .. math::

            \boldsymbol{D}(\boldsymbol{k})
            =
            \begin{pmatrix}
            \boldsymbol{A}(\boldsymbol{k}) & \boldsymbol{B}^{\dagger}(\boldsymbol{k}) \\
            \boldsymbol{B}(\boldsymbol{k}) & \overline{\boldsymbol{A}(-\boldsymbol{k})}
            \end{pmatrix}

        See Also
        --------

        LSWT.A
        LSWT.B

        Examples
        --------

        .. doctest::

            >>> import magnopy
            >>> spinham = magnopy.examples.cubic_ferro_nn()
            >>> lswt = magnopy.LSWT(spinham=spinham, spin_directions=[[0, 0, 1]])
            >>> lswt.GDM(k=[0,0,0.5], relative=True) # doctest: +SKIP
            array([[1.+0.j, 0.-0.j],
                   [0.+0.j, 1.-0.j]])
        """

        k = np.array(k, dtype=float)

        A = self.A(k=k, relative=relative, units=units)
        A_m = self.A(k=-k, relative=relative, units=units)

        B = self.B(k=k, relative=relative, units=units)

        left = np.concatenate((A, np.conjugate(B).T), axis=0)
        right = np.concatenate((B, np.conjugate(A_m)), axis=0)
        gdm = np.concatenate((left, right), axis=1)

        return gdm

    def diagonalize(self, k, relative=False, units="meV"):
        r"""
        Diagonalizes the Hamiltonian for the given ``k`` point.

        Parameters
        ----------

        k : (3,) |array-like|_
            Reciprocal vector.

        relative : bool, default False
            If ``relative=True``, then ``k`` is interpreted as given relative to the
            reciprocal unit cell. Otherwise it is interpreted as given in absolute
            coordinates.

        units : str, default "meV"
            .. versionadded:: 0.3.0

            Units of energy. See :ref:`user-guide_usage_units_magnon-energy-units` for the
            full list of supported units.

        Returns
        -------

        omegas : (M, ) :numpy:`ndarray`
            Array of omegas. Note, that data type is complex. If the ground state is
            correct, then the complex part should be zero.

        delta : float
            Constant energy term that results from diagonalization. Note, that data type
            is complex. If the ground state is correct, then the complex part should be
            zero.

        G : (M, 2M) :numpy:`ndarray`
            Transformation matrix from the original boson operators.

            .. math::

                \begin{pmatrix}
                    b_1(\boldsymbol{k}) \\
                    \dots \\
                    b_M(\boldsymbol{k}) \\
                \end{pmatrix}
                =
                \mathcal{G}
                \begin{pmatrix}
                    a_1(\boldsymbol{k}) \\
                    \dots \\
                    a_M(\boldsymbol{k}) \\
                    a^{\dagger}_1(-\boldsymbol{k}) \\
                    \dots \\
                    a^{\dagger}_M(-\boldsymbol{k}) \\
                \end{pmatrix}

        See Also
        --------

        LSWT.omega
        LSWT.delta
        LSWT.G

        Examples
        --------

        .. doctest::

            >>> import magnopy
            >>> spinham = magnopy.examples.cubic_ferro_nn()
            >>> lswt = magnopy.LSWT(spinham=spinham, spin_directions=[[0, 0, 1]])
            >>> lswt.diagonalize(k=[0, 0, 0.5], relative=True) # doctest: +SKIP
            (array([2.+0.j]), 0j, array([[1.+0.j, 0.+0.j]]))
        """

        GDM = self.GDM(np.array(k, dtype=float), relative=relative)

        # Diagonalize via Colpa's method
        try:
            E, G = solve_via_colpa(GDM)
        except ColpaFailed:
            # Try to diagonalize with suspected Goldstone mode
            try:
                E, G = solve_via_colpa(
                    GDM + (1e-10) * np.eye(GDM.shape[0], dtype=float),
                )
            # Return NaNs if it still fails
            except ColpaFailed:
                # Try to diagonalize for the negative GDMs
                # Note: solve_via_colpa will return positive eigenvalues,
                # so we need to negate them back
                try:
                    E, G = solve_via_colpa(-GDM)
                    E = -E
                except ColpaFailed:
                    return (
                        [np.nan for _ in range(self.M)],
                        np.nan,
                        [[np.nan for _ in range(2 * self.M)] for _ in range(self.M)],
                    )

        # Convert units if necessary
        if units != "meV":
            units = _validated_units(units=units, supported_units=_MAGNON_ENERGY_UNITS)
            tmp_factor = _MAGNON_ENERGY_UNITS["mev"] / _MAGNON_ENERGY_UNITS[units]
            E = E * tmp_factor

        # Factor of two explained in the paper (TODO: add doi after publication)
        energies = E[: self.M] * 2
        transformation_matrices = G[: self.M]

        return (
            energies,  # energies (M)
            complex(0.5 * (np.sum(E[self.M :]) - np.sum(E[: self.M]))),  # delta term
            transformation_matrices,  # transformation matrix (M x 2M)
        )

    def omega(self, k, relative=False, units="meV"):
        r"""
        Computes magnon's eigenenergies at the given ``k`` point.

        Parameters
        ----------

        k : (3,) |array-like|_
            Reciprocal vector.

        relative : bool, default False
            If ``relative=True``, then ``k`` is interpreted as given relative to the
            reciprocal unit cell. Otherwise it is interpreted as given in absolute
            coordinates.

        units : str, default "meV"
            .. versionadded:: 0.3.0

            Units of energy. See :ref:`user-guide_usage_units_magnon-energy-units` for the
            full list of supported units.


        Returns
        -------
        omegas : (M, ) :numpy:`ndarray`
            Array of omegas. Note, that data type is complex. If the ground state is correct,
            then the complex part should be zero.

        See Also
        --------

        LSWT.diagonalize
        LSWT.delta

        Examples
        --------

        .. doctest::

            >>> import magnopy
            >>> spinham = magnopy.examples.cubic_ferro_nn()
            >>> lswt = magnopy.LSWT(spinham=spinham, spin_directions=[[0, 0, 1]])
            >>> lswt.omega(k=[0, 0, 0.5], relative=True)
            array([2.+0.j])
        """

        return self.diagonalize(k=k, relative=relative, units=units)[0]

    def delta(self, k, relative=False, units="meV"):
        r"""
        Computes constant delta term of the diagonalized Hamiltonian.

        .. math::

            \sum_{\boldsymbol{k}}\Delta(\boldsymbol{k})

        Parameters
        ----------

        k : (3,) |array-like|_
            Reciprocal vector.

        relative : bool, default False
            If ``relative=True``, then ``k`` is interpreted as given relative to the
            reciprocal unit cell. Otherwise it is interpreted as given in absolute
            coordinates.

        units : str, default "meV"
            .. versionadded:: 0.3.0

            Units of energy. See :ref:`user-guide_usage_units_magnon-energy-units` for the
            full list of supported units.


        Returns
        -------

        delta : float
            Constant energy term that results from diagonalization. Note, that data type is complex. If the ground state is correct,
            then the complex part should be zero.

        See Also
        --------

        LSWT.diagonalize
        LSWT.omega

        Examples
        --------

        .. doctest::

            >>> import magnopy
            >>> spinham = magnopy.examples.cubic_ferro_nn()
            >>> lswt = magnopy.LSWT(spinham=spinham, spin_directions=[[0, 0, 1]])
            >>> lswt.delta(k=[0, 0, 0.5], relative=True)
            0j
        """
        return self.diagonalize(k=k, relative=relative, units=units)[1]

    def G(self, k, relative=False):
        r"""
        Computes transformation matrix to the new bosonic operators.

        .. math::

            b_{\alpha}(\boldsymbol{k})
            =
            \sum_{\beta}
            (\mathcal{G})_{\alpha, \beta}(\boldsymbol{k})
            \mathcal{A}_{\beta}(\boldsymbol{k})

        Parameters
        ----------

        k : (3,) |array-like|_
            Reciprocal vector

        relative : bool, default False
            If ``relative=True``, then ``k`` is interpreted as given relative to the
            reciprocal unit cell. Otherwise it is interpreted as given in absolute
            coordinates.

        Returns
        -------

        G : (M, 2M) :numpy:`ndarray`
            Transformation matrix from the original boson operators.

        See Also
        --------

        LSWT.diagonalize
        LSWT.omega
        LSWT.delta

        Examples
        --------

        .. doctest::

            >>> import magnopy
            >>> spinham = magnopy.examples.cubic_ferro_nn()
            >>> lswt = magnopy.LSWT(spinham=spinham, spin_directions=[[0, 0, 1]])
            >>> lswt.G(k=[0, 0, 0.5], relative=True)  # doctest: +SKIP
            array([[1.+0.j, 0.+0.j]])
        """
        return self.diagonalize(k=k, relative=relative)[2]

    # REMOVED in v0.2.0. Warning will be removed in March of 2026
    def G_inv(self, *args, **kwargs):
        r"""
        This method was removed in v0.2.0 in favor of LSWT.G.

        Raises
        ------
        DeprecationWarning
            This method was removed in v0.2.0 in favor of :py:meth:`.LSWT.G`.
        """
        raise DeprecationWarning("This method was removed in v0.2.0 in favor of LSWT.G")


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================


import os

import numpy as np

from magnopy._energy import Energy
from magnopy._package_info import logo
from magnopy._spinham._supercell import make_supercell
from magnopy._plotly_engine import PlotlyEngine
from magnopy._constants._icons import ICON_OUT_FILE


# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def optimize_sd(
    spinham,
    supercell=(1, 1, 1),
    magnetic_field=None,
    energy_tolerance=1e-5,
    torque_tolerance=1e-5,
    output_folder="magnopy-results",
    comment=None,
    no_html=False,
    hide_personal_data=False,
) -> None:
    r"""
    Optimizes classical energy of spin Hamiltonian and finds a set of spin directions
    that describes local minima of the energy landscape.

    Progress of calculation is shown in the standard output (``print()``). A bunch of the
    output files is created and saved on the disk inside the ``output_folder``.

    Parameters
    ----------

    spinham : :py:class:`.SpinHamiltonian`
        Spin Hamiltonian object.

    supercell : (3, ) tuple of int, default (1, 1, 1)
        .. versionadded:: 0.2.0

        Tuple with number of repetitions of the unit cell along each lattice vector.
        Spin directions are varied within the supercell during the optimization. By
        default, optimization is done within the unit cell.

    magnetic_field : (3, ) |array-like|_
        Vector of external magnetic field (magnetic flux density, B), given in Tesla.

    energy_tolerance : float, default 1e-5
        Tolerance parameter. Difference between classical energies of two consecutive
        optimization steps.

    torque_tolerance : float, default 1e-5
        Tolerance parameter. Maximum torque among all spins at the current optimization
        step.

    output_folder : str, default "magnopy-results"
        Name for the folder where to save the output files. The folder is created if it
        does not exist.

    comment : str, optional
        Any comment, that will be shown in the standard output right after the magnopy's
        logo.

    no_html : bool, default False
        .. versionadded:: 0.2.0

        Whether to produce .html files. If ``no_html = False``, then |plotly|_ is expected
        to be available.

    hide_personal_data : bool, default False
        .. versionadded:: 0.2.0

        If ``False``, then ``os.path.abspath(pathname)`` is used to show full paths to the output
        and input files. If ``True``, then only ``pathname`` is used.

    Raises
    ------

    ValueError
        If ``len(supercell) != 3``.

    ValueError
        If ``supercell[0] < 1`` or ``supercell[1] < 1`` or ``supercell[2] < 1``.
    """
    ################################################################################
    ##                                  Filenames                                 ##
    ################################################################################
    # Create the output directory if it does not exist
    os.makedirs(output_folder, exist_ok=True)

    def envelope_path(pathname):
        if hide_personal_data:
            return pathname
        else:
            return os.path.abspath(pathname)

    INITIAL_GUESS_TXT = envelope_path(os.path.join(output_folder, "INITIAL_GUESS.txt"))
    SPIN_DIRECTIONS_TXT = envelope_path(
        os.path.join(output_folder, "SPIN_DIRECTIONS.txt")
    )
    SPIN_POSITIONS_TXT = envelope_path(
        os.path.join(output_folder, "SPIN_POSITIONS.txt")
    )
    SPIN_DIRECTIONS_HTML = envelope_path(
        os.path.join(output_folder, "SPIN_DIRECTIONS.html")
    )
    E_0_TXT = envelope_path(os.path.join(output_folder, "E_0.txt"))

    ################################################################################
    ##                           Input data verification                          ##
    ################################################################################

    # Supercell
    supercell = tuple(supercell)
    if len(supercell) != 3:
        raise ValueError(
            f"Expected a tuple of three int for supercell, got {len(supercell)} elements."
        )
    if supercell[0] < 1 or supercell[1] < 1 or supercell[2] < 1:
        raise ValueError(f"Supercell repetitions must be >=1, got {supercell}.")

    ################################################################################
    ##                              Logo and comment                              ##
    ################################################################################
    print(logo(date_time=True, line_length=80))
    if comment is not None:
        print(f"\n{' Comment ':=^80}\n")
        print(comment)

    ################################################################################
    ##                          Optimization parameters                           ##
    ################################################################################
    print(f"\n{' Optimization parameters ':=^80}\n")

    # Tolerance parameters
    print(f"Energy tolerance      : {energy_tolerance:.5e} meV")
    print(f"Torque tolerance      : {torque_tolerance:.5e}")

    # Magnetic field
    if magnetic_field is not None:
        print(
            f"Magnetic flux density : "
            f"|{magnetic_field[0]:.5f}, {magnetic_field[1]:.5f}, {magnetic_field[2]:.5f}|"
            f" = {np.linalg.norm(magnetic_field):.5f} Tesla"
        )
        spinham.add_magnetic_field(B=magnetic_field)
    else:
        print("Magnetic flux density : None")

    # Supercell
    print(
        f"Supercell             : {supercell[0]} x {supercell[1]} x {supercell[2]}",
        end="",
    )
    original_spinham = spinham
    if supercell != (1, 1, 1):
        spinham = make_supercell(spinham=spinham, supercell=supercell)
        print(" (constructed supercell of the Hamiltonian)")
    else:
        print(" (original unit cell of the Hamiltonian)")

    # Initial guess
    initial_guess = np.random.uniform(low=-1, high=1, size=(spinham.M, 3))
    initial_guess = initial_guess / np.linalg.norm(initial_guess, axis=1)[:, np.newaxis]
    np.savetxt(INITIAL_GUESS_TXT, initial_guess, fmt="%12.8f %12.8f %12.8f")
    print(
        f"\nSpin directions of the initial guess are saved in file\n{ICON_OUT_FILE} {INITIAL_GUESS_TXT}"
    )

    ################################################################################
    ##                                Optimization                                ##
    ################################################################################
    print(f"\n{' Start optimization ':=^80}\n")

    energy = Energy(spinham=spinham)
    spin_directions = energy.optimize(
        initial_guess=initial_guess,
        energy_tolerance=energy_tolerance,
        torque_tolerance=torque_tolerance,
        quiet=False,
    )
    print("Optimization is done.")

    ################################################################################
    ##                                 Text output                                ##
    ################################################################################
    print(f"\n{' Output ':=^80}\n")

    # Classical energy
    E_0 = energy.E_0(spin_directions=spin_directions)
    with open(E_0_TXT, "w", encoding="utf-8") as f:
        f.write(f"{E_0:.8f} meV\n")
    print(
        f"Classic energy of optimized state (E_0 = {E_0:.3f} meV) is saved in file\n{ICON_OUT_FILE} {E_0_TXT}"
    )

    # Optimized spin directions
    np.savetxt(SPIN_DIRECTIONS_TXT, spin_directions, fmt="%12.8f %12.8f %12.8f")
    print(
        f"\nOptimized spin directions are saved in file\n{ICON_OUT_FILE} {SPIN_DIRECTIONS_TXT}"
    )

    # Real-space absolute positions of magnetic centers
    positions = np.array(spinham.magnetic_atoms["positions"]) @ spinham.cell
    np.savetxt(SPIN_POSITIONS_TXT, positions, fmt="%12.8f %12.8f %12.8f")
    print(f"\nSpin positions are saved in file\n{ICON_OUT_FILE} {SPIN_POSITIONS_TXT}")

    ################################################################################
    ##                                 HTML output                                ##
    ################################################################################
    if not no_html:
        try:
            pe = PlotlyEngine()

            pe.plot_cell(
                original_spinham.cell,
                color="Black",
                legend_label="Unit cell",
            )

            original_uc_spins_indices = [i for i in range(original_spinham.M)]

            pe.plot_spin_directions(
                positions=positions[original_uc_spins_indices],
                spin_directions=spin_directions[original_uc_spins_indices],
                colors="#A47864",
                legend_label="Spins of the unit cell",
            )

            if supercell != (1, 1, 1):
                other_uc_spins_indices = [
                    i for i in range(original_spinham.M, spinham.M)
                ]
                pe.plot_spin_directions(
                    positions=positions[other_uc_spins_indices],
                    spin_directions=spin_directions[other_uc_spins_indices],
                    colors="#535FCF",
                    legend_label="Spins of other unit cells",
                )

            pe.plot_spin_directions(
                positions=positions,
                spin_directions=initial_guess,
                colors="#0DB00D",
                legend_label="Initial guess",
            )

            pe.save(
                output_name=SPIN_DIRECTIONS_HTML,
                axes_visible=True,
                legend_position="top",
                kwargs_write_html=dict(include_plotlyjs=True, full_html=True),
            )

            print(
                f"\nImage of spin directions is saved in file\n{ICON_OUT_FILE} {SPIN_DIRECTIONS_HTML}"
            )
        except ImportError:
            print(
                "\nCannot produce .html files because plotly is not available.\n"
                "You can install plotly with 'pip install plotly'"
            )
    else:
        print("\nHTML output is disabled by user (no_html=True).\n")

    print(f"\n{' Finished ':=^80}")


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Magnopy - Python package for magnons.
# Copyright (C) 2023-2026 Magnopy Team
#
# e-mail: anry@uv.es, web: magnopy.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================


import numpy as np

from magnopy._parameters._p22 import from_dmi, from_iso
from magnopy._spinham._convention import Convention
from magnopy._spinham._hamiltonian import SpinHamiltonian

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def ivuzjo(N=10, J=10):
    r"""
    Prepares a Hamiltonian from the paper by Ivanov, Uzdin and Jónsson.

    See [1]_ for details. The Hamiltonian is defined as

    .. math::

        \mathcal{H}
        =
        -\dfrac{1}{2}
        \sum_{\mu, \nu}
        J
        \boldsymbol{S}_{\mu}
        \cdot
        \boldsymbol{S}_{\mu+\nu}
        -
        \dfrac{1}{2}
        \sum_{\mu, \nu}
        \dfrac{J}{2}
        \boldsymbol{r}_{\nu}
        \left(
        \boldsymbol{S}_{\mu}
        \times
        \boldsymbol{S}_{\mu+\nu}
        \right)
        +
        \sum_{\mu}
        J
        \boldsymbol{\hat{z}}
        \boldsymbol{S}_{\mu}

    Parameters
    ----------

    N : int, default 10
        Size of the supercell (N x N).

    J : float, default 10
        Value of the isotropic exchange in energy units (meV), sign is *not* ignored.

    Returns
    -------

    spinham : :py:class:`.SpinHamiltonian`
        Spin Hamiltonian (with magnetic field)

    References
    ----------

    .. [1] Ivanov, A.V., Uzdin, V.M. and Jónsson, H., 2021.
        Fast and robust algorithm for energy minimization of spin systems applied
        in an analysis of high temperature spin configurations in terms of skyrmion
        density.
        Computer Physics Communications, 260, p.107749.

    Examples
    --------

    To create an example Hamiltonian use

    .. doctest::

        >>> import magnopy
        >>> spinham = magnopy.examples.ivuzjo()

    """

    D = J / 2

    BOHR_MAGNETON = 0.057883818060  # meV / Tesla

    cell = np.eye(3, dtype=float) * N

    atoms = dict(names=[], positions=[], g_factors=[], spins=[])
    names_to_index = {}
    atom_index = 0
    for i in range(0, N):
        for j in range(0, N):
            atoms["names"].append(f"Fe_{i + 1}_{j + 1}")
            atoms["positions"].append([i + 0.5, j + 0.5, 0])
            atoms["spins"].append(1)
            atoms["g_factors"].append(2)
            names_to_index[f"Fe_{i + 1}_{j + 1}"] = atom_index
            atom_index += 1

    convention = Convention(
        multiple_counting=True, spin_normalized=False, c21=-1, c22=-0.5
    )

    spinham = SpinHamiltonian(cell=cell, atoms=atoms, convention=convention)

    # For each atom add bonds
    for i in range(0, N):
        for j in range(0, N):
            alpha = names_to_index[f"Fe_{i + 1}_{j + 1}"]

            # 1 0 0
            if i == N - 1:
                nu = (1, 0, 0)
                beta = names_to_index[f"Fe_1_{j + 1}"]
            else:
                nu = (0, 0, 0)
                beta = names_to_index[f"Fe_{i + 2}_{j + 1}"]

            parameter = from_iso(iso=J) + from_dmi(dmi=[D, 0, 0])
            spinham.add_22(alpha=alpha, beta=beta, nu=nu, parameter=parameter)

            # 0 1 0
            if j == N - 1:
                nu = (0, 1, 0)
                beta = names_to_index[f"Fe_{i + 1}_1"]
            else:
                nu = (0, 0, 0)
                beta = names_to_index[f"Fe_{i + 1}_{j + 2}"]
            parameter = from_iso(iso=J) + from_dmi(dmi=[0, D, 0])
            spinham.add_22(alpha=alpha, beta=beta, nu=nu, parameter=parameter)

    spinham.add_magnetic_field(B=[0, 0, J / 5 / BOHR_MAGNETON / 2])

    return spinham


def full_ham():
    r"""
    Prepares a Hamiltonian with all parameters being populated.

    Returns
    -------

    spinham : :py:class:`.SpinHamiltonian`
        Spin Hamiltonian.

    Examples
    --------

    To create an example Hamiltonian use

    .. doctest::

        >>> import magnopy
        >>> spinham = magnopy.examples.full_ham()
    """

    cell = np.eye(3, dtype=float)
    atoms = dict(
        names=["Cr1", "Cr2", "Cr3", "Cr4"],
        g_factors=[2, 2, 2, 2],
        spins=[3 / 2, 3 / 2, 3 / 2, 3 / 2],
        positions=[
            [0.0, 0.0, 0.0],
            [0.0, 0.0, 0.5],
            [0.0, 0.5, 0.0],
            [0.0, 0.0, 0.5],
        ],
    )
    convention = Convention(
        spin_normalized=False,
        multiple_counting=True,
        c1=1,
        c21=1,
        c22=1,
        c31=1,
        c32=1,
        c33=1,
        c41=1,
        c421=1,
        c422=1,
        c43=1,
        c44=1,
    )

    spinham = SpinHamiltonian(cell=cell, atoms=atoms, convention=convention)

    ############################################################################
    #                                 One site                                 #
    ############################################################################
    for alpha in range(4):
        spinham.add_1(alpha=alpha, parameter=[0, 0, 1])
        spinham.add_21(alpha=alpha, parameter=[0.5, 1, 1.5])
        spinham.add_31(alpha=alpha, parameter=[-0.4, 0.1, 1])
        spinham.add_41(alpha=alpha, parameter=[0.1, 0.2, 1])

    ############################################################################
    #                                Two sites                                 #
    ############################################################################

    for alpha, beta, nu in [
        [0, 0, (1, 0, 0)],
        [1, 1, (1, 0, 0)],
        [2, 2, (1, 0, 0)],
        [3, 3, (1, 0, 0)],
    ]:
        spinham.add_22(alpha=alpha, beta=beta, nu=nu, parameter=-np.eye(3))
        spinham.add_32(alpha=alpha, beta=beta, nu=nu, parameter=-np.ones((3, 3, 3)))
        spinham.add_421(alpha=alpha, beta=beta, nu=nu, parameter=-np.ones((3, 3, 3, 3)))
        spinham.add_422(alpha=alpha, beta=beta, nu=nu, parameter=-np.ones((3, 3, 3, 3)))

    for alpha, beta in [[0, 1], [1, 2], [2, 3], [3, 1]]:
        spinham.add_22(alpha=alpha, beta=beta, nu=nu, parameter=0.5 * np.eye(3))
        spinham.add_32(
            alpha=alpha, beta=beta, nu=nu, parameter=0.5 * np.ones((3, 3, 3))
        )
        spinham.add_421(
            alpha=alpha, beta=beta, nu=nu, parameter=0.5 * np.ones((3, 3, 3, 3))
        )
        spinham.add_422(
            alpha=alpha, beta=beta, nu=nu, parameter=0.5 * np.ones((3, 3, 3, 3))
        )

    ############################################################################
    #                                Three sites                               #
    ############################################################################
    for alpha, beta, gamma, nu, _lambda in [[0, 1, 2, (0, 0, 0), (0, 0, 0)]]:
        spinham.add_33(
            alpha=alpha,
            beta=beta,
            gamma=gamma,
            nu=nu,
            _lambda=_lambda,
            parameter=0.3 * np.ones((3, 3, 3)),
        )
        spinham.add_43(
            alpha=alpha,
            beta=beta,
            gamma=gamma,
            nu=nu,
            _lambda=_lambda,
            parameter=0.3 * np.ones((3, 3, 3, 3)),
        )
    ############################################################################
    #                                Four sites                                #
    ############################################################################
    for alpha, beta, gamma, epsilon, nu, _lambda, rho in [
        [0, 1, 2, 3, (0, 0, 0), (0, 0, 0), (0, 0, 0)]
    ]:
        spinham.add_44(
            alpha=alpha,
            beta=beta,
            gamma=gamma,
            epsilon=epsilon,
            nu=nu,
            _lambda=_lambda,
            rho=rho,
            parameter=-0.1 * np.ones((3, 3, 3, 3)),
        )

    return spinham


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir
